"""MIT License

Copyright (c) 2021 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging

import discord

from .enums import ApplicationCommandType
from .errors import InvalidArgument
from .utils import get_enum

log = logging.getLogger(__name__)


# OptionType
class Mentionable:
    pass


# Option
class CommandOptionChoice:
    """Represents an application command option choice.

    Attributes
    ----------
    name: str
        The name of the choice. This is visible to the user; max 100 characters.
    value: str
        The name of the choice. This is not visible to the user; max 100 characters.
    """

    def __init__(
        self,
        name: str,
        value: int | str | float,
    ):
        self.name = name
        self.value = value

    @classmethod
    def from_payload(cls, data):
        return cls(name=data["name"], value=data["value"])

    def to_dict(self) -> dict:
        return {"name": self.name, "value": self.value}

    def __eq__(self, other):
        return self.name == other.name and self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)


class CommandOption:
    """Represents an application command option.

    Attributes
    ----------
    description: str
        The description of the option; max 100 characters.
    choices: list[CommandOptionChoice]
        A list of choices for the command to choose from for this option.
    min_value: Union[float, int]
        For option type ``string``, the minimum allowed length; min 0, max 6000
    max_value: Optional[int]
        For option type ``string``, the maximum allowed length; min 1, max 6000
    required: bool
        Whether the option is required. (defaults to ``false``)
    autocomplete: bool
        Whether the option has autocomplete. (defaults to ``false``)
    """

    def __init__(
        self,
        name: str = None,
        option_type: type = None,
        description: str = "No description.",
        choices: list[CommandOptionChoice] = None,
        channel_type: discord.ChannelType | int = None,
        channel_types: list[discord.ChannelType | int] = None,
        min_value: float | int = None,
        max_value: float | int = None,
        required: bool = False,
        autocomplete: bool = False,
    ):
        if choices is None:
            choices = []
        self._name = name
        self._type = option_type
        self.description = description
        self.choices = choices
        self.required = required
        self.autocomplete = autocomplete

        self.parameter_name: str | None = None

        if option_type is not None:
            if (len(choices) > 0 or autocomplete) and not (
                int in option_type.__mro__
                or float in option_type.__mro__
                or str in option_type.__mro__
            ):
                raise TypeError(
                    "choices or autocomplete should only be used in integer, string, and float."
                )

        if len(self.choices) > 0 and self.autocomplete:
            log.warning("autocomplete may not be set to true if choices are present.")

        self._channel_type: list[int] | None = None
        if channel_type is not None or channel_types is not None:
            if option_type is not None:
                if discord.abc.GuildChannel not in option_type.__mro__:
                    raise TypeError(
                        "Channel options can be set only when they are set to a channel type."
                    )

            if channel_types is not None and channel_type is not None:
                raise InvalidArgument()
            elif channel_type is not None:
                if isinstance(channel_type, discord.ChannelType):
                    self._channel_type = [getattr(channel_type, "value", 0)]
                else:
                    self._channel_type = [channel_type]
            elif channel_types is not None:
                self._channel_type = []
                for x in channel_types:
                    if hasattr(x, "value"):
                        self._channel_type.append(x.value)
                    else:
                        self._channel_type.append(x)

        if option_type is not None:
            if min_value is not None and (
                int not in option_type.__mro__ and float not in option_type.__mro__
            ):
                raise TypeError(
                    "min_value can only be called when the parameter types are int and float."
                )
            if max_value is not None and (
                int not in option_type.__mro__ and float not in option_type.__mro__
            ):
                raise TypeError(
                    "max_value can only be called when the parameter types are int and float."
                )
        self.min_value: int | None = min_value
        self.max_value: int | None = max_value

    @classmethod
    def empty_option(cls):
        return cls(None, None)

    @property
    def name(self) -> str:
        """The name of the option"""
        if self._name is None:
            raise ValueError("The name of the option must be defined.")
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def type(self) -> type:
        """The type of the option"""
        if self._type is None:
            raise ValueError("The type of the option must be defined.")
        return self._type

    @property
    def channel_type(self) -> list[discord.ChannelType] | None:
        """A list of channel types that are allowed for this option."""
        if discord.abc.GuildChannel not in self.type.__mro__:
            return
        channel_type = []
        for x in self._channel_type:
            channel_type.append(get_enum(discord.ChannelType, x))
        return channel_type

    @property
    def _get_type_id(self) -> int:
        if ApplicationSubcommand in self.type.__mro__:
            return 1
        elif ApplicationSubcommandGroup in self.type.__mro__:
            return 2
        elif str == self.type:
            return 3
        elif int == self.type:
            return 4
        elif bool == self.type:
            return 5
        elif discord.User in self.type.__mro__:
            return 6
        elif discord.abc.GuildChannel in self.type.__mro__:
            return 7
        elif discord.Role in self.type.__mro__:
            return 8
        elif Mentionable in self.type.__mro__:
            return 9
        elif float == self.type:
            return 10
        elif discord.Attachment == self.type:
            return 11
        raise TypeError(
            "option type invalid (Subcommand and Subcommand Group, please use decorator)"
        )

    def to_dict(self) -> dict:
        data = {
            "name": self.name,
            "description": self.description,
            "type": self._get_type_id,
            "required": self.required,
            "choices": [choice.to_dict() for choice in self.choices],
            "autocomplete": self.autocomplete,
        }
        if self._channel_type is not None:
            data["channel_types"] = self._channel_type
        if self.min_value is not None:
            data["min_value"] = self.min_value
        if self.max_value is not None:
            data["max_value"] = self.max_value
        return data

    def __eq__(self, other):
        default_check = (
            self.name == other.name
            and other._type in self._type.__mro__
            and self.description == other.description
            and self.choices == other.choices
            and self.required == other.required
            and self.autocomplete == other.autocomplete
        )
        if int in self._type.__mro__ or float in self._type.__mro__:
            default_check = (
                default_check
                and self.min_value == other.min_value
                and self.max_value == other.max_value
            )
        elif discord.abc.GuildChannel in self._type.__mro__:
            default_check = default_check and self._channel_type == other._channel_type
        return default_check

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def from_payload(cls, data: dict):
        new_cls = cls.empty_option()
        tp_v = data["type"]
        if tp_v == 1:
            return ApplicationSubcommand.from_payload(data)
        if tp_v == 2:
            return ApplicationSubcommandGroup.from_payload(data)

        for x in ("name", "description", "required", "autocomplete"):
            setattr(new_cls, x, data.get(x))
        # optional--default false
        if new_cls.required is None:
            new_cls.required = False
        if new_cls.autocomplete is None:
            new_cls.autocomplete = False

        new_cls._type = (
            str,
            int,
            bool,
            discord.User,
            discord.abc.GuildChannel,
            discord.Role,
            Mentionable,
            float,
            discord.Attachment,
        )[(tp_v - 3)]
        if new_cls.type == discord.abc.GuildChannel and "channel_types" in data.keys():
            new_cls._channel_type = data.get("channel_types", [])
        if (
            new_cls.type == int or new_cls.type == float
        ) and "min_value" in data.keys():
            new_cls.min_value = data.get("min_value")
        if (
            new_cls.type == int or new_cls.type == float
        ) and "max_value" in data.keys():
            new_cls.max_value = data.get("max_value")

        new_cls.choices = []
        choices = data.get("choices", [])
        if len(choices) > 0:
            for ch in choices:
                new_cls.choices.append(CommandOptionChoice.from_payload(ch))
        return new_cls


class ApplicationSubcommand:
    def __init__(
        self,
        name: str,
        description: str = "No description.",
        options: list[CommandOption] | None = None,
    ):
        self.name = name
        self.description = description
        if options is None:
            options = []
        self.options = options

    @property
    def _get_type_id(self) -> int:
        return 1

    @classmethod
    def from_payload(cls, data: dict):
        new_cls = cls(name=data["name"], description=data["description"])
        if "options" in data:
            new_cls.options = [CommandOption.from_payload(x) for x in data["options"]]
        return new_cls

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "description": self.description,
            "type": self._get_type_id,
            "options": [opt.to_dict() for opt in self.options],
        }

    def __eq__(self, other):
        default_check = (
            self.name == other.name
            and self.description == other.description
            and self.options == other.options
        )
        return default_check

    def __ne__(self, other):
        return not self.__eq__(other)


class ApplicationSubcommandGroup:
    def __init__(
        self,
        name: str,
        options: list[ApplicationSubcommand],
        description: str = "No description.",
    ):
        self.name = name
        self.description = description
        self.options = options

    @property
    def _get_type_id(self) -> int:
        return 2

    @classmethod
    def from_payload(cls, data: dict):
        return cls(
            name=data["name"],
            description=data["description"],
            options=[ApplicationSubcommand.from_payload(x) for x in data["options"]],
        )

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "description": self.description,
            "type": self._get_type_id,
            "options": [opt.to_dict() for opt in self.options],
        }

    def __eq__(self, other):
        default_check = (
            self.name == other.name
            and self.description == other.description
            and self.options == other.options
        )
        return default_check

    def __ne__(self, other):
        return not self.__eq__(other)


class ApplicationCommand:
    """Represents an application command.

    Attributes
    ----------
    id: int
        The id of application command
    name: str
        The name of application command
    description: str
        The description of application command
    version: int
        The version of application command
    type: ApplicationCommandType
        The type of application command
    application_id: ApplicationCommandType
        The application id of application command
    guild_id: Optional[int]
        The guild id of application command for private command
    default_member_permissions: Optional[str]
        The default member permissions that can run this command
    """

    def __init__(
        self,
        name: str,
        description: str = None,
        guild_id: int | None = None,
        command_type: ApplicationCommandType = ApplicationCommandType.CHAT_INPUT,
        default_member_permissions: str = None,
    ):
        self.id: int = 0  # default: None
        self.name: str = name
        self.type: ApplicationCommandType = command_type
        self.application_id: int = 0  # default: None
        self.guild_id: int | None = guild_id
        self.description: str | None = description
        self.default_member_permissions: str | None = default_member_permissions
        self.version: int = 1  # default: None

    @classmethod
    def from_payload(cls, data: dict):
        new_cls = cls(
            name=data["name"],
            description=data["description"],
            default_member_permissions=data.get("default_member_permissions"),
            guild_id=data.get("guild_id"),
        )

        for key in ("id", "application_id", "version"):
            value = data.get(key)
            setattr(new_cls, key, value)
        command_type = data["type"]
        new_cls.type = get_enum(ApplicationCommandType, command_type)
        return new_cls

    def to_register_dict(self) -> dict:
        data = {
            "name": self.name,
            "type": self.type.value,
            "description": self.description,
        }
        if self.default_member_permissions is not None:
            data["default_member_permissions"] = self.default_member_permissions
        return data

    @property
    def is_guild(self) -> bool:
        """Whether a private command(guild command)"""
        return self.guild_id is not None

    def __eq__(self, other):
        description = self.description or ""
        return (
            self.name == other.name
            and self.type == other.type
            and description == other.description
            and self.default_member_permissions == other.default_member_permissions
        )

    def __ne__(self, other):
        return not self.__eq__(other)


class SlashCommand(ApplicationCommand):
    """Represents an application command for ``CHAT_INPUT`` type.

    Attributes
    ----------
    options: list[Union[CommandOption, ApplicationSubcommand, ApplicationSubcommandGroup]]
        A list of options for application command.
    """

    def __init__(
        self,
        options: list[
            CommandOption | ApplicationSubcommandGroup | ApplicationSubcommand
        ] = None,
        **kwargs
    ):
        if options is None:
            options = []
        super().__init__(**kwargs)
        self.type = ApplicationCommandType.CHAT_INPUT
        self.options: list[
            CommandOption | ApplicationSubcommand | ApplicationSubcommandGroup
        ] = options

    def __eq__(self, other):
        return super().__eq__(other) and self.options == other.options

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def from_payload(cls, data: dict):
        new_cls = super().from_payload(data)
        new_cls.options = [
            CommandOption.from_payload(opt) for opt in data.get("options", [])
        ]
        return new_cls

    def to_register_dict(self) -> dict:
        data = super().to_register_dict()
        data["options"] = [opt.to_dict() for opt in self.options]
        return data


class UserCommand(ApplicationCommand):
    """Represents an application command for ``User`` type."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = ApplicationCommandType.USER


class ContextMenu(ApplicationCommand):
    """Represents an application command for ``Context Menu`` type."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = ApplicationCommandType.MESSAGE


command_types = SlashCommand | UserCommand | ContextMenu


def from_payload(data: dict) -> command_types:
    if data["type"] == ApplicationCommandType.CHAT_INPUT:
        _result = SlashCommand.from_payload(data)
    elif data["type"] == ApplicationCommandType.USER:
        _result = UserCommand.from_payload(data)
    elif data["type"] == ApplicationCommandType.MESSAGE:
        _result = ContextMenu.from_payload(data)
    else:
        _result = ApplicationCommand.from_payload(data)
    return _result


# For Decorator
def option(
    name: str = None,
    option_type: type = None,
    description: str = "No description.",
    choices: list[CommandOptionChoice] = None,
    channel_type: discord.ChannelType | int = None,
    channel_types: list[discord.ChannelType | int] = None,
    min_value: float | int = None,
    max_value: float | int = None,
    required: bool = False,
    autocomplete: bool = False,
):
    options = CommandOption(
        name=name,
        option_type=option_type,
        description=description,
        choices=choices,
        channel_type=channel_type,
        channel_types=channel_types,
        min_value=min_value,
        max_value=max_value,
        required=required,
        autocomplete=autocomplete,
    )

    def decorator(func):
        if hasattr(func, "options") or isinstance(func, SlashCommand):
            func.options.append(options)
        else:
            if not hasattr(func, "__command_options__"):
                func.__command_options__ = []
            func.__command_options__.append(options)
        return func

    return decorator



"""MIT License

Copyright (c) 2021 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import functools
import inspect
from typing import Callable

import discord.utils
from discord.ext.commands.errors import *

from .core import BaseCommand
from .interaction import InteractionContext


def check(predicate):
    def decorator(func):
        if isinstance(func, BaseCommand):
            func.checks.append(predicate)
        else:
            if not hasattr(func, "__commands_checks__"):
                func.__commands_checks__ = []

            func.__commands_checks__.append(predicate)

        return func

    if inspect.iscoroutinefunction(predicate):
        decorator.predicate = predicate
    else:

        @functools.wraps(predicate)
        async def wrapper(ctx):
            return predicate(ctx)  # type: ignore

        decorator.predicate = wrapper

    return decorator


def check_any(*checks) -> Callable:
    """A :func:`check` that is added that checks if any of the checks passed
    will pass, i.e. using logical OR.

    If all checks fail then :exc:`.CheckAnyFailure` is raised to signal the failure.
    It inherits from :exc:`.CheckFailure`.

    .. note::

        The ``predicate`` attribute for this function **is** a coroutine.

    Parameters
    ------------
    checks: Callable[[:class:`Context`], :class:`bool`]
        An argument list of checks that have been decorated with
        the :func:`check` decorator.

    Raises
    -------
    TypeError
        A check passed has not been decorated with the :func:`check`
        decorator.

    Examples
    ---------

    Creating a basic check to see if it's the bot owner or
    the server owner:

    .. code-block:: python3

        def is_guild_owner():
            def predicate(ctx):
                return ctx.guild is not None and ctx.guild.owner_id == ctx.author.id
            return commands.check(predicate)

        @bot.command()
        @commands.check_any(commands.is_owner(), is_guild_owner())
        async def only_for_owners(ctx):
            await ctx.send('Hello mister owner!')
    """

    unwrapped = []
    for wrapped in checks:
        try:
            pred = wrapped.predicate
        except AttributeError:
            raise TypeError(
                f"{wrapped!r} must be wrapped by commands.check decorator"
            ) from None
        else:
            unwrapped.append(pred)

    async def predicate(ctx: InteractionContext) -> bool:
        errors = []
        for func in unwrapped:
            try:
                value = await func(ctx)
            except CheckFailure as e:
                errors.append(e)
            else:
                if value:
                    return True
        # if we're here, all checks failed
        raise CheckAnyFailure(unwrapped, errors)

    return check(predicate)


def has_role(item: int | str) -> Callable:
    """A :func:`.check` that is added that checks if the member invoking the
    command has the role specified via the name or ID specified.

    If a string is specified, you must give the exact name of the role, including
    caps and spelling.

    If an integer is specified, you must give the exact snowflake ID of the role.

    If the message is invoked in a private message context then the check will
    return ``False``.

    This check raises one of two special exceptions, :exc:`.MissingRole` if the user
    is missing a role, or :exc:`.NoPrivateMessage` if it is used in a private message.
    Both inherit from :exc:`.CheckFailure`.

    Parameters
    -----------
    item: Union[:class:`int`, :class:`str`]
        The name or ID of the role to check.
    """

    def predicate(ctx: InteractionContext) -> bool:
        if ctx.guild is None:
            raise NoPrivateMessage()

        # ctx.guild is None doesn't narrow ctx.author to Member
        if isinstance(item, int):
            role = discord.utils.get(ctx.author.roles, id=item)  # type: ignore
        else:
            role = discord.utils.get(ctx.author.roles, name=item)  # type: ignore
        if role is None:
            raise MissingRole(item)
        return True

    return check(predicate)


def has_any_role(*items: int | str) -> Callable:
    """A :func:`.check` that is added that checks if the member invoking the
    command has **any** of the roles specified. This means that if they have
    one out of the three roles specified, then this check will return `True`.

    Similar to :func:`.has_role`, the names or IDs passed in must be exact.

    This check raises one of two special exceptions, :exc:`.MissingAnyRole` if the user
    is missing all roles, or :exc:`.NoPrivateMessage` if it is used in a private message.
    Both inherit from :exc:`.CheckFailure`.

    Parameters
    -----------
    items: list[Union[:class:`str`, :class:`int`]]
        An argument list of names or IDs to check that the member has roles wise.

    Example
    --------

    .. code-block:: python3

        @bot.command()
        @commands.has_any_role('Library Devs', 'Moderators', 492212595072434186)
        async def cool(ctx):
            await ctx.send('You are cool indeed')
    """

    def predicate(ctx):
        if ctx.guild is None:
            raise NoPrivateMessage()

        # ctx.guild is None doesn't narrow ctx.author to Member
        getter = functools.partial(discord.utils.get, ctx.author.roles)  # type: ignore
        if any(
            (
                getter(id=item) is not None
                if isinstance(item, int)
                else getter(name=item) is not None
            )
            for item in items
        ):
            return True
        raise MissingAnyRole(list(items))

    return check(predicate)


def bot_has_role(item: int) -> Callable:
    """Similar to :func:`.has_role` except checks if the bot itself has the
    role.

    This check raises one of two special exceptions, :exc:`.BotMissingRole` if the bot
    is missing the role, or :exc:`.NoPrivateMessage` if it is used in a private message.
    Both inherit from :exc:`.CheckFailure`.
    """

    def predicate(ctx):
        if ctx.guild is None:
            raise NoPrivateMessage()

        me = ctx.me
        if isinstance(item, int):
            role = discord.utils.get(me.roles, id=item)
        else:
            role = discord.utils.get(me.roles, name=item)
        if role is None:
            raise BotMissingRole(item)
        return True

    return check(predicate)


def bot_has_any_role(*items: int) -> Callable:
    """Similar to :func:`.has_any_role` except checks if the bot itself has
    any of the roles listed.

    This check raises one of two special exceptions, :exc:`.BotMissingAnyRole` if the bot
    is missing all roles, or :exc:`.NoPrivateMessage` if it is used in a private message.
    Both inherit from :exc:`.CheckFailure`.
    """

    def predicate(ctx):
        if ctx.guild is None:
            raise NoPrivateMessage()

        me = ctx.me
        getter = functools.partial(discord.utils.get, me.roles)
        if any(
            (
                getter(id=item) is not None
                if isinstance(item, int)
                else getter(name=item) is not None
            )
            for item in items
        ):
            return True
        raise BotMissingAnyRole(list(items))

    return check(predicate)


def has_permissions(**perms: bool) -> Callable:
    """A :func:`.check` that is added that checks if the member has all of
    the permissions necessary.

    Note that this check operates on the current channel permissions, not the
    guild wide permissions.

    The permissions passed in must be exactly like the properties shown under
    :class:`.discord.Permissions`.

    This check raises a special exception, :exc:`.MissingPermissions`
    that is inherited from :exc:`.CheckFailure`.

    Parameters
    ------------
    perms
        An argument list of permissions to check for.

    Example
    ---------

    .. code-block:: python3

        @bot.command()
        @commands.has_permissions(manage_messages=True)
        async def test(ctx):
            await ctx.send('You can manage messages.')

    """

    invalid = set(perms) - set(discord.Permissions.VALID_FLAGS)
    if invalid:
        raise TypeError(f"Invalid permission(s): {', '.join(invalid)}")

    def predicate(ctx: InteractionContext) -> bool:
        ch = ctx.channel
        permissions = ch.permissions_for(ctx.author)  # type: ignore

        missing = [
            perm for perm, value in perms.items() if getattr(permissions, perm) != value
        ]

        if not missing:
            return True

        raise MissingPermissions(missing)

    return check(predicate)


def bot_has_permissions(**perms: bool) -> Callable:
    """Similar to :func:`.has_permissions` except checks if the bot itself has
    the permissions listed.

    This check raises a special exception, :exc:`.BotMissingPermissions`
    that is inherited from :exc:`.CheckFailure`.
    """

    invalid = set(perms) - set(discord.Permissions.VALID_FLAGS)
    if invalid:
        raise TypeError(f"Invalid permission(s): {', '.join(invalid)}")

    def predicate(ctx: InteractionContext) -> bool:
        guild = ctx.guild
        me = guild.me if guild is not None else ctx.client.user
        permissions = ctx.channel.permissions_for(me)  # type: ignore

        missing = [
            perm for perm, value in perms.items() if getattr(permissions, perm) != value
        ]

        if not missing:
            return True

        raise BotMissingPermissions(missing)

    return check(predicate)


def has_guild_permissions(**perms: bool) -> Callable:
    """Similar to :func:`.has_permissions`, but operates on guild wide
    permissions instead of the current channel permissions.

    If this check is called in a DM context, it will raise an
    exception, :exc:`.NoPrivateMessage`.
    """

    invalid = set(perms) - set(discord.Permissions.VALID_FLAGS)
    if invalid:
        raise TypeError(f"Invalid permission(s): {', '.join(invalid)}")

    def predicate(ctx: InteractionContext) -> bool:
        if not ctx.guild:
            raise NoPrivateMessage

        permissions = ctx.author.guild_permissions  # type: ignore
        missing = [
            perm for perm, value in perms.items() if getattr(permissions, perm) != value
        ]

        if not missing:
            return True

        raise MissingPermissions(missing)

    return check(predicate)


def bot_has_guild_permissions(**perms: bool) -> Callable:
    """Similar to :func:`.has_guild_permissions`, but checks the bot
    members guild permissions.
    """

    invalid = set(perms) - set(discord.Permissions.VALID_FLAGS)
    if invalid:
        raise TypeError(f"Invalid permission(s): {', '.join(invalid)}")

    def predicate(ctx: InteractionContext) -> bool:
        if not ctx.guild:
            raise NoPrivateMessage

        permissions = ctx.me.guild_permissions  # type: ignore
        missing = [
            perm for perm, value in perms.items() if getattr(permissions, perm) != value
        ]

        if not missing:
            return True

        raise BotMissingPermissions(missing)

    return check(predicate)


def dm_only() -> Callable:
    """A :func:`.check` that indicates this command must only be used in a
    DM context. Only private messages are allowed when
    using the command.

    This check raises a special exception, :exc:`.PrivateMessageOnly`
    that is inherited from :exc:`.CheckFailure`.
    """

    def predicate(ctx: InteractionContext) -> bool:
        if ctx.guild is not None:
            raise PrivateMessageOnly()
        return True

    return check(predicate)


def guild_only() -> Callable:
    """A :func:`.check` that indicates this command must only be used in a
    guild context only. Basically, no private messages are allowed when
    using the command.

    This check raises a special exception, :exc:`.NoPrivateMessage`
    that is inherited from :exc:`.CheckFailure`.
    """

    def predicate(ctx: InteractionContext) -> bool:
        if ctx.guild is None:
            raise NoPrivateMessage()
        return True

    return check(predicate)


def is_owner() -> Callable:
    """A :func:`.check` that checks if the person invoking this command is the
    owner of the bot.

    This is powered by :meth:`.Bot.is_owner`.

    This check raises a special exception, :exc:`.NotOwner` that is derived
    from :exc:`.CheckFailure`.
    """

    async def predicate(ctx: InteractionContext) -> bool:
        if not await ctx.client.is_owner(ctx.author):
            raise NotOwner("You do not own this bot.")
        return True

    return check(predicate)


def is_nsfw() -> Callable:
    """A :func:`.check` that checks if the channel is a NSFW channel.

    This check raises a special exception, :exc:`.NSFWChannelRequired`
    that is derived from :exc:`.CheckFailure`.
    """

    def pred(ctx: InteractionContext) -> bool:
        ch = ctx.channel
        if ctx.guild is None or (
            isinstance(ch, (discord.TextChannel, getattr(discord, "Thread", None)))
            and ch.is_nsfw()
        ):
            return True
        raise NSFWChannelRequired(ch)  # type: ignore

    return check(pred)



from enum import Enum
from typing import Any


class ApplicationCommandType(Enum):
    CHAT_INPUT = 1
    USER = 2
    MESSAGE = 3

    def __eq__(self, other):
        if isinstance(other, ApplicationCommandType):
            return self._value_ == other.value
        return self._value_ == other

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def value(self) -> Any:
        return super(ApplicationCommandType, self).value


class Locale(Enum):
    american_english = "en-US"
    british_english = "en-GB"
    bulgarian = "bg"
    chinese = "zh-CN"
    taiwan_chinese = "zh-TW"
    croatian = "hr"
    czech = "cs"
    indonesian = "id"
    danish = "da"
    dutch = "nl"
    finnish = "fi"
    french = "fr"
    german = "de"
    greek = "el"
    hindi = "hi"
    hungarian = "hu"
    italian = "it"
    japanese = "ja"
    korean = "ko"
    lithuanian = "lt"
    norwegian = "no"
    polish = "pl"
    brazil_portuguese = "pt-BR"
    romanian = "ro"
    russian = "ru"
    spain_spanish = "es-ES"
    swedish = "sv-SE"
    thai = "th"
    turkish = "tr"
    ukrainian = "uk"
    vietnamese = "vi"



"""MIT License

Copyright (c) 2021 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from abc import *
import inspect
from typing import Any
import discord

from .core import BaseCore
from .utils import get_enum


class Components(metaclass=ABCMeta):
    TYPE: int | None = None

    def __init__(self, components_type: discord.ComponentType):
        self.type = components_type

    @abstractmethod
    def to_dict(self) -> dict:
        pass

    @classmethod
    @abstractmethod
    def from_dict(cls, payload: dict) -> dict:
        pass

    def __eq__(self, other):
        return self.type == other.type

    def __ne__(self, other):
        return not self.__eq__(other)


class SelectOption:
    """Represents a select menu’s option.

    Attributes
    ----------
    label : str
        The label of option. This is visible to the user; max 100 characters.
    value : str
        The value of option. This is not visible to the user; max 100 characters.
    description : Optional[str]
        Additional description of the option; max 100 characters
    emoji : Optional[discord.PartialEmoji]
        The emoji of the option, if available.
        If ``emoji`` attributes is used as a dict, key requires an id, name, and animation.
    default : Optional[bool]
        Will show this option as selected by default
    """

    def __init__(
        self,
        label: str,
        value: str,
        description: str | None = None,
        emoji: discord.PartialEmoji | dict = None,
        default: bool = False,
    ):
        self.label = label
        self.value = value
        self.description = description
        self.emoji = emoji
        self.default = default

    def __str__(self) -> str:
        if self.emoji:
            base = f"{self.emoji} {self.label}"
        else:
            base = self.label

        if self.description:
            return f"{base}\n{self.description}"
        return base

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self) -> dict:
        data = {"label": self.label, "value": self.value}

        if self.description is not None:
            data["description"] = self.description
        if self.emoji is not None:
            data["emoji"] = (
                self.emoji.to_dict()
                if isinstance(self.emoji, discord.PartialEmoji)
                else self.emoji
            )
        if self.default is not None:
            data["default"] = self.default

        return data

    @classmethod
    def from_dict(cls, payload: dict):
        label: str = payload.get("label")
        value: str = payload.get("value")
        description: str | None = payload.get("description")
        emoji: discord.PartialEmoji = payload.get("emoji")
        default: bool = payload.get("default", False)
        return cls(
            label=label,
            value=value,
            description=description,
            emoji=emoji,
            default=default,
        )


class ActionRow(Components):
    """Represents an Action Row.

    An Action Row is a non-interactive container component for other types of components.
    This can contain up to five other Components.
    And, It cannot contain an Action Row.

    Attributes
    ----------
    components : list[Components]
        Contains up to five Components.
    """

    TYPE = 1

    def __init__(self, components: list[Components] = None):
        super().__init__(components_type=discord.ComponentType.action_row)

        self.components: list = components

    def to_dict(self) -> dict:
        return {"type": 1, "components": self.components}

    def to_all_dict(self) -> dict:
        return {"type": 1, "components": [i.to_dict() for i in self.components]}

    @classmethod
    def from_dict(cls, payload: dict):
        components = payload.get("components")
        return cls(components=components)

    @classmethod
    def from_payload(cls, payload: dict):
        components = from_payload(payload.get("components"))
        return cls(components=components)

    def __eq__(self, other):
        return self.components == other.components and super().__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)


class Button(Components):
    """Represents a Button.

    Attributes
    ----------
    style : Union[:class:`discord.ButtonStyle`, int]
        The style of the button.
    label : Optional[str]
        The label of the button.
    emoji : Optional[Union[:class:`discord.PartialEmoji`, str, dict]]
        The emoji of the button, if available.
        If ``emoji`` attributes is used as a dict, key requires an id, name, and animation.
    custom_id: Optional[str]
        The custom_id of button. This is not visible to the user; max 100 characters.
    url: Optional[str]
        The URl of button for link-style(5).
    disabled: Optional[bool]
        Whether the button is disabled. This default is ``false``
    """

    TYPE = 2

    def __init__(
        self,
        style: int | discord.ButtonStyle,
        label: str = None,
        emoji: discord.PartialEmoji | str | dict = None,
        custom_id: str = None,
        url: str = None,
        disabled: bool = None,
    ):
        super().__init__(components_type=discord.ComponentType.button)

        # Accepting a button's style as an int will be disabled.
        if isinstance(style, int):
            style = get_enum(discord.ButtonStyle, style)
        self.style = style
        self.label = label
        self.emoji = emoji
        self.custom_id = custom_id
        self.url = url
        self.disabled = disabled

    def to_dict(self) -> dict:
        base = {"type": 2, "style": int(self.style)}

        if self.label is not None:
            base["label"] = self.label
        if self.emoji is not None and isinstance(self.emoji, discord.PartialEmoji):
            base["emoji"] = self.emoji.to_dict()
        elif self.emoji is not None and isinstance(self.emoji, str):
            base["emoji"] = {"name": self.emoji}
        elif self.emoji is not None:
            base["emoji"] = self.emoji

        if 0 < int(self.style) < 5 and self.custom_id is not None:
            base["custom_id"] = self.custom_id
        if int(self.style) == 5 and self.url is not None:
            base["url"] = self.url

        if self.disabled is not None:
            base["disabled"] = self.disabled

        return base

    @classmethod
    def from_dict(cls, payload: dict):
        style = get_enum(discord.ButtonStyle, payload["style"])
        label = payload.get("label")
        emoji = payload.get("emoji")
        custom_id = payload.get("custom_id")
        url = payload.get("url")
        disabled = payload.get("disabled", False)
        return cls(
            style=style,
            label=label,
            emoji=emoji,
            custom_id=custom_id,
            url=url,
            disabled=disabled,
        )

    def __eq__(self, other):
        return self.custom_id == other.custom_id and super().__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)


class Selection(Components):
    """Represents a Select Menu.

    Attributes
    ----------
    custom_id: str
        The custom_id of select menu. This is not visible to the user; max 100 characters.
    options : list[:class:`.Option`]
        The options of the select menu.
    disabled: Optional[bool]
        Whether the select menu is disabled. This default is ``false``
    placeholder : Optional[str]
        The placeholder of the select menu.
    min_values : Optional[int]
        Minimum number of items that can be chosen (defaults to 1); min 0, max 25
    max_values: Optional[int]
        Maximum number of items that can be chosen (defaults to 1); max 25
    """

    TYPE = 3

    def __init__(
        self,
        custom_id: str,
        options: list[dict | SelectOption],
        disabled: bool = False,
        placeholder: str = None,
        min_values: int = None,
        max_values: int = None,
    ):
        super().__init__(components_type=discord.ComponentType.select)

        self.disabled = disabled
        self.custom_id = custom_id
        self.options = options
        self.placeholder = placeholder
        self.min_values = min_values
        self.max_values = max_values

    def to_dict(self) -> dict:
        base = {
            "type": 3,
            "custom_id": self.custom_id,
            "disabled": self.disabled,
            "options": [
                option.to_dict() if isinstance(option, SelectOption) else option
                for option in self.options
            ],
        }
        if self.placeholder is not None:
            base["placeholder"] = self.placeholder
        if self.min_values is not None:
            base["min_values"] = self.min_values
        if self.max_values is not None:
            base["max_values"] = self.max_values

        return base

    def __eq__(self, other):
        return (
            self.custom_id == other.custom_id
            and super().__eq__(other)
            and self.options == other.options
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def from_dict(cls, payload: dict):
        custom_id = payload["custom_id"]
        options = [SelectOption.from_dict(x) for x in payload.get("options", [])]
        placeholder = payload.get("placeholder")
        min_values = payload.get("min_values")
        max_values = payload.get("max_values")
        return cls(
            custom_id=custom_id,
            options=options,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
        )


class TextInput(Components):
    """Represents a TextInput.

    Notes
    -----
    Text inputs are an interactive component that render on modals.

    Attributes
    ----------
    custom_id: str
        The custom_id of text input. This is not visible to the user; max 100 characters.
    style : Union[:class:`discord.TextStyle`, int]
        The style of the text input.
    label: str
        The label of text input. This is visible to the user; max 45 characters
    placeholder : Optional[str]
        The placeholder of the text input.
    min_length : Optional[int]
        Minimum input length for a text input; min 0, max 4000
    max_length: Optional[int]
        Maximum input length for a text input; min 1, max 4000
    required: Optional[bool]
        Whether this input text is required to be filled (defaults to ``true``)
    value: Optional[str]
        Pre-filled value for this component; max 4000 characters
    """

    TYPE = 4

    def __init__(
        self,
        custom_id: str,
        style: int,
        label: str,
        min_length: int | None = None,
        max_length: int | None = None,
        required: bool = False,
        value: str | None = None,
        placeholder: str | None = None,
    ):
        super().__init__(components_type=discord.ComponentType.text_input)

        self.custom_id = custom_id
        self.style = style
        self.label = label
        self.placeholder = placeholder
        self.min_length = min_length
        self.max_length = max_length
        self.required = required
        self.value = value

    def __eq__(self, other):
        return self.custom_id == other.custom_id and super().__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self) -> dict:
        base = {
            "type": 4,
            "custom_id": self.custom_id,
            "style": self.style,
            "label": self.label,
            "required": self.required,
        }
        if self.placeholder is not None:
            base["placeholder"] = self.placeholder
        if self.min_length is not None:
            base["min_length"] = self.min_length
        if self.max_length is not None:
            base["max_length"] = self.max_length
        if self.value is not None:
            base["value"] = self.value

        return base

    @classmethod
    def from_dict(cls, payload: dict):
        custom_id = payload["custom_id"]
        style = payload.get("style")
        label = payload.get("label")
        placeholder = payload.get("placeholder")
        min_length = payload.get("min_length")
        max_length = payload.get("max_length")
        required = payload.get("required", False)
        value = payload.get("value")
        return cls(
            custom_id=custom_id,
            style=style,
            label=label,
            placeholder=placeholder,
            min_length=min_length,
            max_length=max_length,
            required=required,
            value=value,
        )


def from_payload(
    payload: list[dict[str, Any]]
) -> list[ActionRow | Button | Selection | TextInput]:
    components = []

    for i in payload:
        if i.get("type") == 1:
            components.append(ActionRow.from_payload(i))
        elif i.get("type") == 2:
            components.append(Button.from_dict(i))
        elif i.get("type") == 3:
            components.append(Selection.from_dict(i))
        elif i.get("type") == 4:
            components.append(TextInput.from_dict(i))
    return components


# For Decorator
class DetectComponent(BaseCore):
    def __init__(
        self, func, custom_id, component_type: type[Components] = None, checks=None
    ):
        self.custom_id = custom_id
        self.type = component_type
        self.func = func
        super().__init__(func=func, checks=checks)

    @property
    def type_id(self) -> int | None:
        if isinstance(self.type, Components):
            return self.type.TYPE
        return


def detect_component(
    cls: classmethod = None,
    custom_id: str = None,
    component_type: type[Components] = None,
    checks=None,
):
    """A decorator that transforms a function into a :class:`.DetectComponent`

    Parameters
    ----------
    cls
        The class to construct with.
        You usually don't change ``cls``.
    custom_id: Optional[str]
        The custom id for detect component.
    component_type: Type[Components]
        The component_type for detect component.
    checks
    """
    if cls is None:
        cls = DetectComponent

    def decorator(func):
        _function = func
        if isinstance(func, staticmethod):
            _function = func.__func__

        if not inspect.iscoroutinefunction(_function):
            raise TypeError("Detect Component function must be a coroutine function.")

        new_cls = cls(
            func=_function,
            custom_id=custom_id or _function.__name__,
            component_type=component_type,
            checks=checks,
        )
        return new_cls

    return decorator



"""MIT License

Copyright (c) 2021 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import discord
from discord.http import Route
from discord.utils import MISSING
from typing import Any, NamedTuple
from collections.abc import Sequence


from .components import ActionRow, Button, Selection
from .utils import to_json


class MultipartParameters(NamedTuple):
    payload: dict[str, Any] | None
    multipart: list[dict[str, Any]] | None
    files: Sequence[discord.File] | None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, traceback):
        if self.files:
            for file in self.files:
                file.close()


def handler_message_parameter(
    content: str | None = MISSING,
    *,
    tts: bool = False,
    embed: discord.Embed = MISSING,
    embeds: Sequence[discord.Embed] = MISSING,
    nonce: int | str | None = None,
    flags: discord.MessageFlags = MISSING,
    file: discord.File = MISSING,
    files: Sequence[discord.File] = MISSING,
    allowed_mentions: discord.AllowedMentions = MISSING,
    attachments: Sequence[discord.Attachment | discord.File] = MISSING,
    components: list[ActionRow | Button | Selection] = MISSING,
    reference: discord.MessageReference | discord.PartialMessage = MISSING,
    previous_allowed_mentions: discord.AllowedMentions | None = None,
    mention_author: bool = None,
    stickers: list[discord.Sticker | int] = MISSING,
):
    if files is not MISSING and file is not MISSING:
        raise TypeError("Cannot mix file and files keyword arguments.")
    if embeds is not MISSING and embed is not MISSING:
        raise TypeError("Cannot mix embed and embeds keyword arguments.")

    if file is not MISSING:
        files = [file]

    payload = {"tts": tts}
    if embeds is not MISSING:
        if len(embeds) > 10:
            raise ValueError("embeds has a maximum of 10 elements.")
        payload["embeds"] = [e.to_dict() for e in embeds]

    if embed is not MISSING:
        if embed is None:
            payload["embeds"] = []
        else:
            payload["embeds"] = [embed.to_dict()]

    if content is not MISSING:
        if content is not None:
            payload["content"] = str(content)
        else:
            payload["content"] = None

    if components is not MISSING:
        if components is not None:
            payload["components"] = [
                i.to_all_dict() if isinstance(i, ActionRow) else i.to_dict()
                for i in components
            ]
        else:
            payload["components"] = []

    if nonce is not None:
        payload["nonce"] = str(nonce)

    if reference is not MISSING:
        payload["message_reference"] = reference

    if stickers is not MISSING:
        if stickers is not None:
            payload["sticker_ids"] = [
                sticker.id if isinstance(stickers, discord.Sticker) else stickers
                for sticker in stickers
            ]
        else:
            payload["sticker_ids"] = []

    if flags is not MISSING:
        payload["flags"] = flags.value

    if allowed_mentions:
        if previous_allowed_mentions is not None:
            payload["allowed_mentions"] = previous_allowed_mentions.merge(
                allowed_mentions
            ).to_dict()
        else:
            payload["allowed_mentions"] = allowed_mentions.to_dict()
    elif previous_allowed_mentions is not None:
        payload["allowed_mentions"] = previous_allowed_mentions.to_dict()

    if mention_author is not None:
        if "allowed_mentions" not in payload:
            payload["allowed_mentions"] = discord.AllowedMentions().to_dict()
        payload["allowed_mentions"]["replied_user"] = mention_author

    if attachments is MISSING:
        attachments = files
    else:
        files = [a for a in attachments if isinstance(a, discord.File)]

    if attachments is not MISSING:
        file_index = 0
        attachments_payload = []
        for attachment in attachments:
            if isinstance(attachment, discord.File):
                attachments_payload.append(attachment.to_dict(file_index))
                file_index += 1
            else:
                attachments_payload.append(attachment.to_dict(file_index))

        payload["attachments"] = attachments_payload

    multipart = []
    if files:
        multipart.append({"name": "payload_json", "value": to_json(payload)})
        payload = None
        for index, file in enumerate(files):
            multipart.append(
                {
                    "name": f"files[{index}]",
                    "value": file.fp,
                    "filename": file.filename,
                    "content_type": "application/octet-stream",
                }
            )

    return MultipartParameters(payload=payload, multipart=multipart, files=files)


class InteractionData(NamedTuple):
    id: int
    application_id: str
    token: str


class InteractionHTTPClient:
    def __init__(self, http: discord.http.HTTPClient):
        self.http = http

    # Interaction Response
    async def post_initial_response(
        self, data: InteractionData, payload: dict[str, Any]
    ):
        r = Route(
            "POST", "/interactions/{id}/{token}/callback", id=data.id, token=data.token
        )
        return await self.http.request(r, json=payload)

    async def get_initial_response(self, data: InteractionData):
        r = Route(
            "GET",
            "/webhooks/{id}/{token}/messages/@original",
            id=data.application_id,
            token=data.token,
        )
        return await self.http.request(r)

    async def edit_initial_response(
        self,
        data: InteractionData,
        payload: dict[str, Any] = None,
        form: list[dict[str, Any]] = None,
        files: Sequence[discord.File] | None = MISSING,
    ):
        if form is None:
            form = []

        if len(form) > 0:
            return await self.edit_followup(
                message_id="@original", form=form, files=files, data=data
            )
        return await self.edit_followup(
            message_id="@original", payload=payload, data=data
        )

    async def delete_initial_response(self, data: InteractionData):
        await self.delete_followup(message_id="@original", data=data)

    # Interaction Response (Followup)
    async def post_followup(
        self,
        data: InteractionData,
        payload: dict[str, Any] = None,
        form: list[dict[str, Any]] = None,
        files: Sequence[discord.File] | None = MISSING,
    ):
        if form is None:
            form = []
        r = Route(
            "POST", "/webhooks/{id}/{token}", id=data.application_id, token=data.token
        )
        if len(form) > 0:
            return await self.http.request(r, form=form, files=files)
        return await self.http.request(r, json=payload)

    async def edit_followup(
        self,
        data: InteractionData,
        message_id,
        payload: dict[str, Any] = None,
        form: list[dict[str, Any]] = None,
        files: Sequence[discord.File] | None = MISSING,
    ):
        if form is None:
            form = []
        r = Route(
            "PATCH",
            "/webhooks/{id}/{token}/messages/{message_id}",
            id=data.application_id,
            token=data.token,
            message_id=message_id,
        )
        if len(form) > 0:
            return await self.http.request(r, form=form, files=files)
        return await self.http.request(r, json=payload)

    async def delete_followup(self, data: InteractionData, message_id):
        r = Route(
            "DELETE",
            "/webhooks/{id}/{token}/messages/{message_id}",
            id=data.application_id,
            token=data.token,
            message_id=message_id,
        )
        await self.http.request(r)



"""MIT License

Copyright (c) 2021 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import inspect
import logging
import discord

try:
    import orjson
except ModuleNotFoundError:
    HAS_ORJSON = False
else:
    HAS_ORJSON = True


if HAS_ORJSON:
    _from_json = orjson.loads
else:
    _from_json = json.loads


channel_types = [
    discord.TextChannel,
    discord.VoiceChannel,
    discord.DMChannel,
    discord.StageChannel,
    discord.GroupChannel,
    discord.CategoryChannel,
    discord.ForumChannel,
]


try:
    from deprecated import deprecated
except ModuleNotFoundError:

    def deprecated(version: str, reason: str):
        def decorator(func):
            logging.getLogger("discord.ext.interaction.deprecated").warning(
                f"This has been disabled due to {reason} in version {version}."
            )
            return func

        return decorator


def get_as_snowflake(data, key):
    try:
        value = data[key]
    except KeyError:
        return None
    else:
        return value and int(value)


def get_enum(cls, val):
    enum_val = [i for i in cls if i.value == val]
    if len(enum_val) == 0:
        return val
    return enum_val[0]


def to_json(obj):
    return json.dumps(obj, separators=(",", ":"), ensure_ascii=False)


async def async_all(gen, *, check=inspect.isawaitable):
    for elem in gen:
        if check(elem):
            elem = await elem
        if not elem:
            return False
    return True



"""MIT License

Copyright (c) 2021 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import asyncio
import copy
import importlib
import importlib.machinery
import importlib.util
import inspect
import logging
import os
import sys
import types
import zlib
from typing import Any

import discord.http
from discord.gateway import DiscordWebSocket
from discord.state import ConnectionState

from ._types import CoroutineFunction, UserCheck, _Coroutine, T
from .commands import ApplicationCommand, from_payload, command_types
from .components import DetectComponent
from .core import SubCommand, SubCommandGroup, BaseCommand, decorator_command_types
from .enums import ApplicationCommandType
from .errors import *
from .http import InteractionHTTPClient
from .interaction import (
    ApplicationContext,
    ComponentsContext,
    AutocompleteContext,
    ModalContext,
)
from .message import Message
from .utils import _from_json, async_all

log = logging.getLogger()


class ClientBase:
    _connection: discord.state.ConnectionState
    loop: asyncio.AbstractEventLoop
    http: discord.http.HTTPClient

    def __init__(
        self,
        global_sync_command: bool = False,
        intents: discord.Intents = discord.Intents.default(),
        **options,
    ):
        if discord.version_info.major >= 2:
            options["enable_debug_events"] = True
        super().__init__(intents=intents, **options)
        self.global_sync_command = global_sync_command

        self.__buffer = bytearray()
        self.__zlib = zlib.decompressobj()

        self._application_id_value = None
        self._interactions_of_group = []
        self._interactions: list[dict[str, decorator_command_types]] = [
            dict(),
            dict(),
            dict(),
        ]
        self._fetch_interactions: list[dict[str, ApplicationCommand]] | None = None

        self._detect_components: dict[str, list[DetectComponent]] = dict()

        self.__sync_command_before_ready_register = []
        self.__sync_command_before_ready_popping = []

        self._checks: list[UserCheck] = []

        self._deferred_components: dict[str, list] = dict()
        self._deferred_global_components: list = list()

        self._multiple_setup_hook: list[CoroutineFunction] = list()

        self.extra_events: dict[str, list[CoroutineFunction]] = dict()
        self.__extensions: dict[str, types.ModuleType] = dict()

        self.interaction_http = InteractionHTTPClient(self.http)

        self.extra_events["on_ready"] = [self.on_ready]

    def dispatch(self, event_name: str, /, *args: Any, **kwargs: Any) -> None:
        # super() will resolve to Client
        super().dispatch(event_name, *args, **kwargs)  # type: ignore
        ev = "on_" + event_name
        for event in self.extra_events.get(ev, []):
            self._schedule_event(event, ev, *args, **kwargs)  # type: ignore

    async def on_ready(self):
        await self._sync_command_task()

    async def setup_hook(self):
        await super(ClientBase, self).setup_hook()
        for func in self._multiple_setup_hook:
            await func()
        return

    # Command
    async def register_command(self, command: ApplicationCommand):
        """
        Register application commands with Discord Bot.

        Parameters
        ----------
        command: ApplicationCommand
            Application Command to register with discord bot.
        """
        command_ids = await self._fetch_command_cached()
        if command.name in command_ids[command.type.value - 1]:
            raise CommandRegistrationError(command.name)

        return await self.http.upsert_global_command(
            await self._application_id(), payload=command.to_register_dict()
        )

    async def _find_command(
        self, command: ApplicationCommand, command_id: int
    ) -> int | None:
        if command_id is None and command.id is None:
            command_ids = await self._fetch_command_cached()
            if command.name not in command_ids[command.type.value - 1]:
                raise CommandNotFound(f'Command "{command.name}" is not found')

            command_id = command_ids[command.type.value - 1][command.name].id
        return command_id

    async def edit_command(
        self, command: ApplicationCommand, command_id: int | None = None
    ):
        """Edit application commands with Discord Bot.

        Parameters
        ----------
        command : ApplicationCommand
            Application Command to edit with discord bot.
        command_id : Optional[int]
            Edit application command's id
        """
        command_id = await self._find_command(command, command_id)
        return await self.http.edit_global_command(
            await self._application_id(),
            command_id=command_id or command.id,
            payload=command.to_register_dict(),
        )

    async def delete_command(self, command: ApplicationCommand, command_id: int = None):
        """Delete application commands with Discord Bot.

        Parameters
        ----------
        command : ApplicationCommand
            Application Command to delete with discord bot.
        command_id : Optional[int]
            Delete application command's id
        """
        command_id = await self._find_command(command, command_id)
        return await self.http.delete_global_command(
            await self._application_id(), command_id=command_id or command.id
        )

    # Listener
    def add_listener(self, func: CoroutineFunction, name: str = None):
        """Add a listener for discord bot event call.

        Parameters
        ----------
        func
            Coroutine function to call when an event is called in Discord Bot
        name
        """
        name = name or func.__name__

        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Listeners must be coroutines")

        if name in self.extra_events:
            self.extra_events[name].append(func)
        else:
            self.extra_events[name] = [func]

    def remove_listener(self, func: CoroutineFunction, name: str = None):
        """Delete a listener for discord bot event call.

        Parameters
        ----------
        func
            Coroutine function to delete from bot listeners
        name : Optional[str]
            Name of the listener for the invoked event.
        """
        if name in self.extra_events:
            try:
                self.extra_events[name].remove(func)
            except ValueError:
                pass

    def listen(self, name: str = None):
        """A decorator that add a listener to discord bot.

        Parameters
        ----------
        name : Optional[str]
            Name of the listener for the invoked event.
        """

        def decorator(func: CoroutineFunction) -> CoroutineFunction:
            self.add_listener(func, name)
            return func

        return decorator

    # Multiple setup_hook
    def add_setup_hook(self, func: CoroutineFunction):
        """Add a setup_hook

        Unlike the setup_hook in class:discord.Client,
        when discord bot ready to setup, it calls all the registered corutine functions at the same time.

        Parameters
        ----------
        func
            Coroutine function to call when setting up the loop
        """
        self._multiple_setup_hook.append(func)
        return

    def remove_setup_hook(self, func: CoroutineFunction):
        """Delete a setup_hook function from multi_setup_hook.

        Parameters
        ----------
        func
            Coroutine function to call when setting up the loop
        """
        try:
            self._multiple_setup_hook.append(func)
        except ValueError:
            pass
        return

    def multiple_setup_hook(self):
        """A decorator that add a setup_hook to discord bot."""

        def decorator(func: CoroutineFunction) -> CoroutineFunction:
            self.add_setup_hook(func)
            return func

        return decorator

    # Application ID (from store data)
    async def _application_id(self):
        if self._application_id_value is None:
            application_info: discord.AppInfo = await self.application_info()
            self._application_id_value = application_info.id
        return self._application_id_value

    async def fetch_commands(self) -> list[dict[str, command_types]]:
        """Fetch and update all application commands registered in discord to discord bot."""
        data = await self.http.get_global_commands(await self._application_id())
        result = [{}, {}, {}]  # list order: [
        #     ApplicationCommandType.CHAT_INPUT,
        #     ApplicationCommandType.USER,
        #     ApplicationCommandType.MESSAGE
        # ]
        for x in data:
            _x = from_payload(x)
            result[_x.type.value - 1][_x.name] = _x
        self._fetch_interactions = result
        return result

    async def fetch_command(
        self,
        command_id: int,
        use_cached: bool = False,
        command_type: ApplicationCommandType = None,
    ) -> command_types:
        """Fetches and updates the application commands specified in the discord to the client.

        Parameters
        ----------
        command_id : int
            ID of application command to load
        use_cached : Optional[bool]
            Look for it in preloaded application command list.
            If application command list is not found in preloaded list, it is automatically updated.
        command_type : Optional[ApplicationCommandType]
            Application Command Type. Default value is ``CHAT_INPUT``

        Returns
        -------
            Retrieves application command data registered in discord.
        """
        if command_type is None:
            command_type = ApplicationCommandType.CHAT_INPUT

        if use_cached:
            cached_command_list = await self._fetch_command_cached()
            if command_id in cached_command_list[command_type.value - 1].keys():
                return cached_command_list[command_type.value - 1][command_id]
        data = await self.http.get_global_command(
            application_id=await self._application_id(), command_id=command_id
        )
        _result = from_payload(data)
        if (
            use_cached
            and command_id
            not in self._fetch_interactions[command_type.value - 1].keys()
        ):
            self._fetch_interactions[command_type.value - 1][_result.name] = _result
        return _result

    async def _fetch_command_cached(self) -> list[dict[str, ApplicationCommand]]:
        if self._fetch_interactions is None:
            await self.fetch_commands()
        return self._fetch_interactions

    async def _sync_command(self, command: command_types):
        await self.wait_until_ready()
        fetch_data = await self._fetch_command_cached()
        if command.name in fetch_data[command.type.value - 1].keys():
            command_id = fetch_data[command.type.value - 1][command.name].id
            if command.name in self._interactions[command.type.value - 1]:
                self._interactions[command.type.value - 1][command.name].id = command_id
            if fetch_data[command.type.value - 1][command.name] != command:
                await self.edit_command(command=command)
        else:
            await self.register_command(command)
        return

    async def _sync_command_popping(self, command: command_types):
        await self.wait_until_ready()
        fetch_data = await self._fetch_command_cached()
        if command.name in fetch_data[command.type.value - 1].keys():
            command_id = fetch_data[command.type.value - 1][command.name].id
            await self.delete_command(command, command_id=command_id)
        return

    def _load_from_module_spec(
        self, spec: importlib.machinery.ModuleSpec, key: str, **kwargs
    ) -> None:
        # precondition: key not in self.__extensions
        lib = importlib.util.module_from_spec(spec)
        sys.modules[key] = lib
        try:
            spec.loader.exec_module(lib)  # type: ignore
        except Exception as e:
            del sys.modules[key]
            raise ExtensionFailed(key, e)

        try:
            setup = getattr(lib, "setup")
        except AttributeError:
            del sys.modules[key]
            raise NoEntryPointError(key)

        try:
            setup(self, **kwargs)
        except Exception as e:
            del sys.modules[key]
            raise ExtensionFailed(key, e) from e
        else:
            self.__extensions[key] = lib

    @staticmethod
    def _resolve_name(name: str, package: str | None) -> str:
        try:
            return importlib.util.resolve_name(name, package)
        except ImportError:
            raise ExtensionNotFound(name)

    def load_extension(
        self, name: str, *, package: str | None = None, **kwargs
    ) -> None:
        """Loads an extension.

        An extension is a python module that contains commands, cogs, or listeners.

        An extension must have a global function,
         setup defined as the entry point on what to do when the extension is loaded.
        This entry point must have a single argument, the bot.

        Parameters
        ----------
        name : str
            The extension name to load. It must be dot separated like regular Python imports if accessing a submodule.
            e.g. `foo.test` if you want to import `foo/test.py`.
        package : Optional[str]
            The package name to resolve relative imports with.
            This is required when loading an extension using a relative path,
            e.g `.foo.test.` Defaults to None.
        kwargs :
            Additional argument values to pass to predefined setup function.
        """
        name = self._resolve_name(name, package)
        if name in self.__extensions:
            raise ExtensionAlreadyLoaded(name)

        spec = importlib.util.find_spec(name)
        if spec is None:
            raise ExtensionNotFound(name)

        self._load_from_module_spec(spec, name, **kwargs)
        return

    def load_extensions(self, package: str, directory: str = None, **kwargs) -> None:
        """Fetches all extensions in the specified folder.
        They must have the setup function configured.

        Parameters
        ----------
        package :
            The package name to resolve relative imports with.
        directory : Optional[str]
            The path to the package address to read
            Usually takes an absolute path value.
        kwargs :
            Additional argument values to pass to predefined setup function.
        """
        if directory is not None:
            _package = os.path.join(directory, package)
        else:
            _package = package
        cogs = [
            "{0}.{1}".format(package, file[:-3])
            for file in os.listdir(_package)
            if file.endswith(".py")
        ]

        for cog in cogs:
            self.load_extension(cog, **kwargs)
        return

    async def _sync_command_task(self):
        if len(self.__sync_command_before_ready_register) != 0:
            log.info(
                f"Register registered commands before bot is ready. List: "
                f"{', '.join([x.name for x in self.__sync_command_before_ready_register])}"
            )
        while len(self.__sync_command_before_ready_register) != 0:
            command = self.__sync_command_before_ready_register.pop()
            await self._sync_command(command=command)
        while len(self.__sync_command_before_ready_popping) != 0:
            command = self.__sync_command_before_ready_popping.pop()
            await self._sync_command_popping(command=command)

        if self.global_sync_command:
            log.info(
                "global_sync_command is activated. Delete unregistered commands on client."
            )
            popping_data = await self._fetch_command_cached()
            for index in range(3):
                for already_cmd in self._interactions[index]:
                    if already_cmd in popping_data[index]:
                        del popping_data[index][already_cmd]

                for cmd in popping_data[index].values():
                    await self.delete_command(cmd)

    def add_detect_component(self, detect_component: DetectComponent, _parent=None):
        """Register for the detect_component event.

        When the user presses a specific button or makes a selection,
        the appropriate detect_component event for the custom_id is called.

        Parameters
        ----------
        detect_component: DetectComponent
            Coroutine functions that are called when a button is pressed or a selection is made.
        _parent
            These parameters is used for cog.
        """
        name = detect_component.custom_id
        if _parent is not None:
            detect_component.cog = _parent

        if name in self._detect_components:
            self._detect_components[name].append(detect_component)
        else:
            self._detect_components[name] = [detect_component]
        return

    def get_detect_component(self):
        """Get all of detect_components"""
        return self._detect_components

    def remove_detect_component(
        self, custom_id: str, detect_component: DetectComponent = None
    ):
        """Remove detect_component function.

        If detect_component is empty, all detect_components corresponding to the custom_id are deleted.

        Parameters
        ----------
        custom_id : str
            Custom_id value to detect. This can be a Button or a Selection.
        detect_component : Optional[DetectComponent]
            The detect_component function to delete, which defaults to None.
            If detect_component is None, delete all detect_components corresponding to custom_id.
        """
        if detect_component is None:
            self._detect_components.pop(custom_id)
        else:
            for i, x in enumerate(self._detect_components[custom_id]):
                if x == detect_component:
                    self._detect_components[custom_id].pop(i)
                    break
        return

    def add_interaction(
        self, command: decorator_command_types, sync_command: bool = None, _parent=None
    ):
        """Add interaction command to discord bot

        If sync_command is True,
        it will be synchronized with Discord and
        registered to a command in the Discord application if one does not exist.

        Parameters
        ----------
        command : Union[Command, MemberCommand, ContextMenuCommand]
            Application commands to register
        sync_command : Optional[bool]
            It synchronizes with Discord.
            If it's not in the Discord application command list,
            it will automatically add the command to the register_command function.

            The default value is None,
            which means that if it is None,
            it will follow the synchronization status ``global_sync_command`` attribute set by discord bot client.
        _parent
            These parameters is used for cog.

        Warnings
        --------
            If a command synchronization request is made before the Discord Bot is ready,
            it will wait and synchronize the command when it is ready.
        """
        if sync_command is None:
            sync_command = self.global_sync_command

        if command.name in self._interactions[command.type.value - 1]:
            raise CommandRegistrationError(command.name)

        if _parent is not None:
            command.cog = _parent

            # Add cog to subcommand in command's option.
            if getattr(command, "is_subcommand", False):
                for index, sub_command in enumerate(command.options):
                    if not isinstance(sub_command, (SubCommand, SubCommandGroup)):
                        continue
                    command.options[index].cog = command.cog

        if command.type == ApplicationCommandType.CHAT_INPUT:
            command.set_signature_option()
        self._interactions[command.type.value - 1][command.name] = command

        if sync_command:
            if self.is_ready():
                self._schedule_event(
                    self._sync_command, "sync_command", command=command
                )
            else:
                self.__sync_command_before_ready_register.append(command)
        return

    def get_interaction(self):
        """Get all interaction command included Application Command, User Command and Context Menu Command"""
        result = []
        for x in self._interactions:
            result += x.values()
        return result

    def delete_interaction(self, command: command_types, sync_command: bool = None):
        """Remove interaction command from discord bot

        If sync_command is True
        it will be synchronized with Discord and
        Delete the command if it exists in Discord.

        Parameters
        ----------
        command : Union[Command, MemberCommand, ContextMenuCommand]
            Application commands to delete
        sync_command : Optional[bool]
            It synchronizes with Discord.
            If it's not in the Discord application command list,
            it will automatically add the command to the delete_command function.

            The default value is None,
            which means that if it is None,
            it will follow the synchronization status ``global_sync_command`` attribute set by discord bot client.

        Warnings
        --------
            If a command synchronization request is made before the Discord Bot is ready,
            it will wait and synchronize the command when it is ready.
        """
        if sync_command is None:
            sync_command = self.global_sync_command

        if command.name not in self._interactions[command.type.value - 1]:
            raise CommandNotFound(f'Command "{command.name}" is not found')

        self._interactions[command.type.value - 1].pop(command.name)

        if sync_command:
            if self.is_ready():
                self._schedule_event(
                    self._sync_command_popping, "sync_command", command=command
                )
            else:
                self.__sync_command_before_ready_popping.append(command)

    def add_interaction_cog(self, interaction_cog: T):
        """Add a "cog" to the bot.

        A cog is a class that has its own event listeners, detect_components and commands.

        Parameters
        ----------
        interaction_cog : type
            The cog to register to the bot.
        """
        self._interactions_of_group.append(interaction_cog)
        for func, attr in inspect.getmembers(interaction_cog):
            if isinstance(attr, BaseCommand):
                attr: decorator_command_types
                self.add_interaction(attr, attr.sync_command, interaction_cog)
            elif isinstance(attr, DetectComponent):
                self.add_detect_component(attr, interaction_cog)
            elif inspect.iscoroutinefunction(attr):
                if hasattr(attr, "__cog_listener__") and hasattr(
                    attr, "__cog_listener_names__"
                ):
                    if not attr.__cog_listener__:
                        continue
                    for name in attr.__cog_listener_names__:
                        self.add_listener(attr, name=name)
        return

    # Socket Decoding
    async def on_socket_raw_receive(self, msg):
        if type(msg) is bytes:
            self.__buffer.extend(msg)
            if len(msg) < 4 or msg[-4:] != b"\x00\x00\xff\xff":
                return
            try:
                msg = self.__zlib.decompress(self.__buffer)
            except zlib.error as error:
                # zlib.error: Error -3 while decompressing data: invalid stored block lengths
                log.debug(
                    "zlib.error: {0}\npayload data: {1}".format(
                        [str(arg) for arg in error.args], msg
                    )
                )
                log.warning("zlib.error: Client will reset zlib decompress object")
                self.__zlib = zlib.decompressobj()
                msg = self.__zlib.decompress(self.__buffer)
            msg = msg.decode("utf-8")
            self.__buffer = bytearray()
        payload = _from_json(msg)

        data = payload.get("d", {})
        t = payload.get("t", "")
        op = payload.get("op", "")

        if op != DiscordWebSocket.DISPATCH:
            return
        self.dispatch("payload_receive", payload=payload)

        state: ConnectionState = self._connection
        if t == "INTERACTION_CREATE":
            state.dispatch("interaction_create", payload)
            if data.get("type") == 2:
                result = ApplicationContext(data, self)
                if len(self._interactions[result.application_type - 1]) != 0:
                    state.dispatch("interaction_command", result)
            elif data.get("type") == 3:
                result = ComponentsContext(data, self)
                await self.process_components(result)
                state.dispatch("components", result)
            elif data.get("type") == 4:
                result = AutocompleteContext(data, self)
                state.dispatch("autocomplete", result)
            elif data.get("type") == 5:
                result = ModalContext(data, self)
                state.dispatch("modal", result)
            return
        elif t == "MESSAGE_CREATE":
            channel, _ = getattr(state, "_get_guild_channel")(data)
            message = Message(state=state, data=data, channel=channel)
            state.dispatch("interaction_message", message)
            return

    # Application Context
    async def process_interaction(self, ctx: ApplicationContext):
        _state: ConnectionState = self._connection
        command = self._interactions[ctx.application_type - 1].get(ctx.name)
        if command is None:
            return

        _state.dispatch("command", ctx)

        try:
            func = ctx.function = command
            if command.cog is not None:
                ctx.parents = command.cog

            if await self.can_run(ctx):
                _option = {}
                if ctx.application_type == ApplicationCommandType.CHAT_INPUT.value:
                    _option = ctx.options
                    is_subcommand = getattr(command, "is_subcommand", False)
                    if is_subcommand:
                        options = command.options
                        if "subcommand_group" in ctx.options:
                            sub_command_group = ctx.options["subcommand_group"]
                            for opt in command.options:
                                if (
                                    isinstance(opt, SubCommandGroup)
                                    and sub_command_group.name == opt.name
                                ):
                                    options = opt.options
                                    break
                        sub_command = ctx.options["subcommand"]
                        for opt in options:
                            if opt.name == sub_command.name and isinstance(
                                opt, SubCommand
                            ):
                                ctx.function = func = opt
                                ctx.parents = None
                                if opt.cog is not None:
                                    ctx.parents = opt.cog
                                _option = sub_command.options
                                break
                if await func.can_run(ctx):
                    if ctx.application_type == ApplicationCommandType.CHAT_INPUT.value:
                        for f_opt in copy.copy(_option).keys():
                            for opt in func.options:
                                if opt.name == f_opt and opt.parameter_name != f_opt:
                                    _option[opt.parameter_name] = _option.pop(f_opt)
                                elif opt.name == f_opt:
                                    break
                        await func.callback(ctx, **_option)
                    else:
                        await func.callback(ctx)
                else:
                    raise CheckFailure("The check functions for command failed.")
            else:
                raise CheckFailure("The global check once functions failed.")
        except Exception as error:
            if isinstance(error, CheckFailure):
                _state.dispatch("command_permission_error", ctx, error)
            _state.dispatch("interaction_command_error", ctx, error)
            raise error
        else:
            _state.dispatch("command_complete", ctx)
        return

    async def on_interaction_command(self, ctx: ApplicationContext):
        await self.process_interaction(ctx)
        return

    # Components
    def wait_for_component(
        self, custom_id: str, check=None, timeout=None
    ) -> _Coroutine[ComponentsContext]:
        """Wait for the component with the specified custom_id to be sent.

        The ``timeout`` parameter is passed to :func:`asyncio.wait_for()`.
        Note that by default it will not time out; if it does,
        it will propagate an :exc:`asyncio.TimeoutError`, which is provided for ease of use.

        Parameters
        ----------
        custom_id : str
            Custom ID for detect component
        check : Optional[Callable[..., bool]]
            A predicate to check what to wait for.
            The arguments must meet the parameters of the event being waited for.
        timeout : Optional[float]
            The number of seconds to wait before timing out and raising :exc:`asyncio.TimeoutError`.

        Returns
        -------
            Returns a `ComponentsContext` that satisfies the custom_id.
        """
        future = self.loop.create_future()
        if check is None:

            def _check(_: ComponentsContext):
                return True

            check = _check

        ev = custom_id.lower()
        try:
            listeners = self._deferred_components[ev]
        except KeyError:
            listeners = []
            self._deferred_components[ev] = listeners

        listeners.append((future, check, False))
        return asyncio.wait_for(future, timeout)

    def wait_for_global_component(
        self, check=None, timeout=None
    ) -> _Coroutine[ComponentsContext]:
        """Unconstrained by custom_id, waits for any component_id.

        The ``timeout`` parameter is passed to :func:`asyncio.wait_for()`.
        Note that by default it will not time out; if it does,
        it will propagate an :exc:`asyncio.TimeoutError`, which is provided for ease of use.

        Parameters
        ----------
        check : Optional[Callable[..., bool]]
            A predicate to check what to wait for.
            The arguments must meet the parameters of the event being waited for.
        timeout : Optional[float]
            The number of seconds to wait before timing out and raising :exc:`asyncio.TimeoutError`.

        Returns
        -------
            Returns a :class:`ComponentContext` that satisfies the condition on `check`.
        """
        future = self.loop.create_future()
        if check is None:

            def _check(_: ComponentsContext):
                return True

            check = _check

        self._deferred_global_components.append((future, check, True))
        return asyncio.wait_for(future, timeout)

    async def can_run(self, ctx: ApplicationContext | ComponentsContext) -> bool:
        data = self._checks
        if len(data) == 0:
            return True

        return await async_all(f(ctx) for f in data)  # type: ignore

    async def process_components(self, component: ComponentsContext):
        _state: ConnectionState = self._connection

        detect_component = self._detect_components.get(component.custom_id)
        if detect_component is None:
            detect_component = []
        active_component = []
        for _component in detect_component:
            if (
                _component.type_id == component.component_type
                or _component.type is None
            ):
                try:
                    if await self.can_run(component):
                        if await _component.can_run(component):
                            await _component.callback(component)
                    else:
                        raise CheckFailure("The global check once functions failed.")
                except Exception as error:
                    if isinstance(error, CheckFailure):
                        _state.dispatch("component_permission_error", component, error)
                    _state.dispatch("component_error", component, error)
                else:
                    _state.dispatch("component_complete", component)
                    active_component.append(component)

        listeners = copy.copy(self._deferred_global_components)
        listeners += copy.copy(self._deferred_components.get(component.custom_id, []))
        detect_component_wait_for = []
        if len(listeners) > 0:
            removed = []
            global_removed = []
            for index, (future, check, global_component) in enumerate(listeners):
                if future.cancelled():
                    if global_component:
                        global_removed.append(index)
                    else:
                        removed.append(index)
                    continue

                try:
                    result = check(component)
                except Exception as exc:
                    future.set_exception(exc)
                    if global_component:
                        global_removed.append(index)
                    else:
                        removed.append(index)
                else:
                    if result:
                        detect_component_wait_for.append(component)
                        future.set_result(component)
                        if global_component:
                            global_removed.append(index)
                        else:
                            removed.append(index)

            if len(removed) == len(listeners):
                self._deferred_components.pop(component.custom_id)
            else:
                for idx in reversed(removed):
                    self._deferred_components.get(component.custom_id, []).pop(idx)

            for idx in reversed(global_removed):
                self._deferred_global_components.pop(idx)
        if len(detect_component_wait_for) == 0 and len(active_component) == 0:
            _state.dispatch("components_cancelled", component)
        return


class Client(ClientBase, discord.Client):
    pass


class AutoShardedClient(ClientBase, discord.AutoShardedClient):
    pass



"""MIT License

Copyright (c) 2021 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging
from collections.abc import Sequence

import discord
from discord.channel import _channel_factory
from discord.state import ConnectionState
from discord.utils import MISSING

from .commands import CommandOptionChoice
from .components import (
    ActionRow,
    Button,
    Selection,
    Components,
    from_payload,
    TextInput,
)
from .enums import Locale
from .errors import AlreadyDeferred
from .http import InteractionHTTPClient, InteractionData, handler_message_parameter
from .message import Message
from .utils import get_as_snowflake, channel_types, get_enum

log = logging.getLogger()


class InteractionContext:
    """Represents an interaction context from Discord

    Attributes
    ----------
    client: discord.Client
        The client that is handling this interaction.
    id: int
        The interaction's ID.
    version: int
        The interaction's version.
    type: discord.InteractionType
        The interaction's type
    token: str
        The interaction's token for response.
    application: int
        The interaction's application id
    author: Union[discord.Member, discord.User]
        The user or member that sent the interaction.
    created_at: datetime.datetime
        When the interaction was created.
    locale: Locale
        Selected language of the invoking user
    guild_locale: Locale
        Selected language of the invoking guild
    deferred: bool
        When ``defer`` called, deferred becomes ``True``
    responded: bool
        Whether the Interaction responded through send, defer, update and defer_update
    """

    def __init__(self, payload: dict, client):
        self.client = client
        self.id: int = get_as_snowflake(payload, "id")
        self.version: int = payload.get("version")
        self.type: discord.InteractionType = get_enum(
            discord.InteractionType, payload.get("type")
        )
        self.token = payload.get("token")
        self.application = get_as_snowflake(payload, "application_id")

        self._state: ConnectionState = getattr(client, "_connection")

        self.guild_id = payload.get("guild_id")
        self.channel_id = payload.get("channel_id")

        if self.guild is not None:
            member = payload.get("member")
            self.author = discord.Member(
                data=member, state=self._state, guild=self.guild
            )
        else:
            user = payload.get("user")
            self.author = discord.User(data=user, state=self._state)
        self.created_at = discord.utils.snowflake_time(self.id)
        self.locale = get_enum(Locale, payload.get("locale"))
        self.guild_locale = get_enum(Locale, payload.get("guild_locale"))

        self.deferred = False
        self.responded = False

        self.data = InteractionData(
            token=self.token, id=self.id, application_id=self.application
        )
        if hasattr(self.client, "interaction_http"):
            self.http = self.client.interaction_http
        else:
            self.http = InteractionHTTPClient(http=self.client.http)

    @property
    def guild(self) -> discord.Guild | None:
        """The guild the interaction was sent from."""
        if self.guild_id is not None:
            return self.client.get_guild(int(self.guild_id))
        return

    @property
    def channel(self) -> discord.TextChannel | discord.PartialMessageable | None:
        """The channel the interaction was sent from."""
        if self.channel_id is not None:
            if self.guild is not None:
                channel = self.guild.get_channel(int(self.channel_id))
            else:
                tp = (
                    discord.ChannelType.text
                    if self.guild_id is not None
                    else discord.ChannelType.private
                )
                channel = discord.PartialMessageable(
                    state=self._state, id=self.channel_id, type=tp
                )
            return channel
        return

    @staticmethod
    def _get_payload(
        content=None,
        tts: bool = False,
        embed=None,
        hidden: bool = False,
        allowed_mentions=None,
        components=None,
    ) -> dict:
        payload = {}
        if content:
            payload["content"] = content
        if tts:
            payload["tts"] = tts
        if embed:
            payload["embeds"] = embed
        if allowed_mentions:
            payload["allowed_mentions"] = allowed_mentions
        if hidden:
            payload["flags"] = 1 << 6
        if components is not None:
            payload["components"] = components
        return payload

    @property
    def voice_client(self) -> discord.VoiceClient | None:
        if self.guild is None:
            return None
        return self.guild.voice_client

    async def defer(self, hidden: bool = False):
        """Defers the interaction response.
        If it takes a long time to respond, use defer to let the user wait.

        Parameters
        ----------
        hidden: bool
            Indicates whether to hide delayed messages.
        """
        if self.deferred:
            raise AlreadyDeferred

        base = {"type": 5}
        if hidden:
            base["data"] = {"flags": 64}

        await self.http.post_initial_response(payload=base, data=self.data)
        self.deferred = True
        return

    async def send(
        self,
        content: str | None = MISSING,
        *,
        tts: bool = False,
        embed: discord.Embed = MISSING,
        embeds: list[discord.Embed] = MISSING,
        file: discord.File = MISSING,
        files: list[discord.File] = MISSING,
        hidden: bool = False,
        allowed_mentions: discord.AllowedMentions = None,
        suppress_embeds: bool = False,
        components: list[ActionRow | Button | Selection] = None,
    ):
        """Responds to this interaction by sending a message.

        Parameters
        ----------
        content: Optional[str]
            The content of the message to send
        tts: bool
            Indicates if the message should be sent using text-to-speech.
        embed: Optional[discord.Embed]
            The rich embed for the content to send. This cannot be mixed with ``embeds`` parameter.
        embeds: Optional[list[discord.Embed]]
            A list of embeds to send with the content. Maximum of 10. This cannot be mixed with the ``embed`` parameter.
        file: Optional[discord.File]
            The file to upload.
        files: Optional[list[discord.File]]
            A list of files to upload. Must be a maximum of 10.
        hidden: bool
            Indicates whether to hide delayed messages.
        allowed_mentions: discord.AllowedMentions
            Controls the mentions being processed in this message.
        suppress_embeds: bool
            Whether to suppress embeds for the message. This sends the message without any embeds if set to True.
        components: list[Components]
            The component to send with the message
        """
        if suppress_embeds or hidden:
            flags = discord.MessageFlags(
                ephemeral=hidden,
                suppress_embeds=(
                    suppress_embeds if suppress_embeds and not self.responded else False
                ),
            )
        else:
            flags = MISSING

        params = handler_message_parameter(
            content=content,
            tts=tts,
            embed=embed,
            embeds=embeds,
            file=file,
            files=files,
            allowed_mentions=allowed_mentions,
            components=components,
            flags=flags,
        )

        if not self.responded:
            if (files is not MISSING or file is not MISSING) and not self.deferred:
                await self.defer(hidden=hidden)

            if self.deferred:
                resp = await self.http.edit_initial_response(
                    payload=params.payload,
                    form=params.multipart,
                    files=params.files,
                    data=self.data,
                )
                self.deferred = False
            else:
                await self.http.post_initial_response(
                    payload={"type": 4, "data": params.payload}, data=self.data
                )
                resp = await self.http.get_initial_response(data=self.data)
            self.responded = True
        else:
            resp = await self.http.post_followup(
                payload=params.payload,
                form=params.multipart,
                files=params.files,
                data=self.data,
            )
        ret = Message(state=self._state, channel=self.channel, data=resp)
        return ret

    async def edit(
        self,
        message_id="@original",
        content=None,
        *,
        embed: discord.Embed = MISSING,
        embeds: list[discord.Embed] = MISSING,
        attachments: Sequence[discord.Attachment | discord.File] = MISSING,
        allowed_mentions: discord.AllowedMentions = MISSING,
        components: list[ActionRow | Button | Selection] = MISSING,
    ):
        """Responds to this interaction by editing the original message or followed message.

        Parameters
        ----------
        message_id: str
            The id of message to edit.
            It can replace followed message and respond message.
            Default message id is ``@original``
        content: Optional[str]
            The content of the message to edit
        embed: Optional[discord.Embed]
            The rich embed for the content to edit. This cannot be mixed with ``embeds`` parameter.
        embeds: Optional[list[discord.Embed]]
            A list of embeds to edit with the content. Maximum of 10. This cannot be mixed with the ``embed`` parameter.
        attachments: Optional[list[discord.File]]
            A list of attachments to keep in the message as well as new files to upload.
            If ``[]`` is passed then all attachments are removed.
        allowed_mentions: discord.AllowedMentions
            Controls the mentions being processed in this message.
        components: list[Components]
            The component to edit with the message
        """
        params = handler_message_parameter(
            content=content,
            embed=embed,
            embeds=embeds,
            previous_allowed_mentions=self._state.allowed_mentions,
            attachments=attachments,
            allowed_mentions=allowed_mentions,
            components=components,
        )

        if message_id == "@original":
            resp = await self.http.edit_initial_response(
                payload=params.payload,
                form=params.multipart,
                files=params.files,
                data=self.data,
            )
        else:
            resp = await self.http.edit_followup(
                message_id=message_id,
                payload=params.payload,
                form=params.multipart,
                files=params.files,
                data=self.data,
            )
        ret = Message(state=self._state, channel=self.channel, data=resp)
        if self.deferred:
            self.deferred = False
        return ret

    async def delete(self, message_id="@original"):
        if message_id == "@original":
            await self.http.delete_initial_response(data=self.data)
        else:
            await self.http.delete_followup(message_id=message_id, data=self.data)
        return


class ModalPossible(InteractionContext):
    async def modal(self, custom_id: str, title: str, components: list[Components]):
        """Respond to this interaction by sending a modal.

        Parameters
        ----------
        custom_id: str
            The ID of modal that gets received during an interaction.
        title: str
            The title of the modal.
        components: list[Components]
            A list of component included in the modal.
            Only input-text and action row are used.
        """
        self.responded = True
        payload = {
            "type": 9,
            "data": {
                "custom_id": custom_id,
                "title": title,
                "components": [
                    i.to_all_dict() if isinstance(i, ActionRow) else i.to_dict()
                    for i in components
                ],
            },
        }
        return await self.http.post_initial_response(data=self.data, payload=payload)


class BaseApplicationContext(ModalPossible):
    def __init__(self, payload: dict, client):
        super().__init__(payload, client)
        self._state: ConnectionState = getattr(client, "_connection")
        self._from_data(payload.get("data", {}))

    def _from_data(self, data):
        self.target_id = data.get("target_id")
        self._resolved = data.get("resolved", {})

    def target(self, target_type, target_id: int = None):
        if target_id is None:
            target_id = self.target_id

        if target_type == "message" and "messages" in self._resolved:
            resolved = self._resolved.get("messages", {})
            data = Message(
                state=self._state,
                channel=self.channel,
                data=resolved.get(str(target_id)),
            )
            return data
        elif (
            target_type == "members"
            and "members" in self._resolved
            and self.guild_id is not None
        ):
            resolved = self._resolved.get("members", {})
            member_data = resolved.get(str(target_id), {})

            # USER DATA INJECT!
            user_resolved = self._resolved.get("users", {})
            user_data = user_resolved.get(str(target_id), {})
            member_data["user"] = user_data

            data = discord.Member(data=member_data, state=self._state, guild=self.guild)
            return data
        elif target_type == "users" and "users" in self._resolved:
            resolved = self._resolved.get("users", {})
            data = discord.User(
                data=resolved.get(str(target_id), {}), state=self._state
            )
            return data
        elif (
            target_type == "roles"
            and "roles" in self._resolved
            and self.guild is not None
        ):
            resolved = self._resolved.get("roles", {})
            data = discord.Role(
                data=resolved.get(str(target_id), {}),
                state=self._state,
                guild=self.guild,
            )
            return data
        elif target_type == "channels" and "channels" in self._resolved:
            resolved = self._resolved.get("channels", {})
            data = resolved.get(str(target_id))
            if data is None:
                return

            factory, ch_type = _channel_factory(data["type"])
            if factory is None:
                raise discord.InvalidData(
                    "Unknown channel type {type} for channel ID {id}.".format_map(data)
                )

            if ch_type in (discord.ChannelType.group, discord.ChannelType.private):
                channel = factory(me=self.client.user, data=data, state=self._state)
            else:
                if "position" not in data:
                    data["position"] = None

                guild_id = int(data.get("guild_id") or self.guild_id)
                guild = self.client.get_guild(guild_id) or discord.Object(id=guild_id)
                channel = factory(guild=guild, state=self._state, data=data)

            return channel


class SubcommandContext(BaseApplicationContext):
    """Represents a Discord interaction subcommand response.

    Attributes
    ----------
    name: str
        The name of the interaction.
    options: dict[str, Any]
        All response options
    """

    def __init__(self, original_payload: dict, payload: dict, client):
        super().__init__(original_payload, client)
        self.name = payload.get("name")
        self.options = {}

        for option in payload.get("options", []):
            key = option.get("name")
            value = option.get("value")
            option_type = option.get("type")

            if option_type == 1:
                self.options["subcommand"] = SubcommandContext(
                    original_payload, option, client
                )
            elif option_type == 3:
                self.options[key]: str = value
            elif option_type == 4:
                self.options[key]: int = value
            elif option_type == 5:
                self.options[key] = bool(value)
            elif option_type == 6:
                if self.guild is not None:
                    self.options[key] = self.guild.get_member(value) or self.target(
                        "members", target_id=value
                    )
                else:
                    self.options[key] = client.get_user(value) or self.target(
                        "users", target_id=value
                    )
            elif option_type == 7:
                self.options[key]: channel_types | None = client.get_channel(
                    value
                ) or self.target("channels", target_id=value)
            elif option_type == 8:
                self.options[key]: discord.Role | None = self.guild.get_role(
                    value
                ) or self.target("roles", target_id=value)
            elif option_type == 10:
                self.options[key]: float = float(value)
            elif option_type == 11:
                state: ConnectionState = getattr(client, "_connection")
                self.options[key]: discord.Attachment = discord.Attachment(
                    data=value, state=state
                )
            else:
                self.options[key] = value


class ApplicationContext(BaseApplicationContext):
    """Represents a Discord interaction response.

    Attributes
    ----------
    name: str
        The name of the interaction.
    options: dict[str, Any]
        All response options
    """

    def __init__(self, payload: dict, client):
        super().__init__(payload, client)
        self.type = discord.InteractionType.application_command
        data = payload.get("data", {})

        self.function = None
        self.parents = None

        self.application_type = data.get("type")
        self.name = data.get("name")
        if self.application_type == 1:
            self.options = {}
            self.option_focused = []
            for option in data.get("options", []):
                key = option.get("name")
                value = option.get("value")
                option_type = option.get("type")
                focused = option.get("focused", False)
                if focused:
                    self.option_focused.append(key)

                if option_type == 1:
                    self.options["subcommand"] = SubcommandContext(
                        payload, option, client
                    )
                elif option_type == 2:
                    self.options["subcommand_group"] = SubcommandContext(
                        payload, option, client
                    )
                elif option_type == 3:
                    self.options[key]: str = value
                elif option_type == 4:
                    self.options[key]: int = value
                elif option_type == 5:
                    self.options[key] = bool(value)
                elif option_type == 6:
                    if self.guild is not None:
                        self.options[key] = self.guild.get_member(value) or self.target(
                            "members", target_id=value
                        )
                    else:
                        self.options[key] = client.get_user(value) or self.target(
                            "users", target_id=value
                        )
                elif option_type == 7:
                    self.options[key]: channel_types | None = client.get_channel(
                        value
                    ) or self.target("channels", target_id=value)
                elif option_type == 8:
                    self.options[key]: discord.Role | None = self.guild.get_role(
                        value
                    ) or self.target("roles", target_id=value)
                elif option_type == 10:
                    self.options[key]: float = float(value)
                elif option_type == 11:
                    state: ConnectionState = getattr(client, "_connection")
                    self.options[key]: discord.Attachment = discord.Attachment(
                        data=value, state=state
                    )
                else:
                    self.options[key] = value

        self.command_id = data.get("id")

    @property
    def content(self):
        if self.application_type == 1:
            options = [str(self.options[i]) for i in self.options.keys()]
            return f"/{self.name} {' '.join(options)}"
        else:
            return f"/{self.name}"

    @property
    def is_context(self) -> bool:
        return self.function is not None


class ComponentsContext(ModalPossible):
    """A responded context consisting of a modal.

    Attributes
    ----------
    custom_id: str
        The ID of component.
    component_type: int
        The component type of component.
    values: list[str]
        If the component type is select menu, it will contain the items selected by the user.
    message: Message
        The original message of component.
    """

    def __init__(self, payload: dict, client):
        super().__init__(payload, client)
        self.type = discord.InteractionType.component
        data = payload.get("data", {})

        self.custom_id = data.get("custom_id")
        self.component_type = data.get("component_type")
        if self.component_type == 3:
            self.values: list[str] = data.get("values")
        else:
            self.values: list[str] = []

        self.message = Message(
            state=self._state, channel=self.channel, data=payload.get("message", {})
        )

    async def defer_update(self, hidden: bool = False):
        """Defers the interaction response to updates.
        If it takes a long time to update, use defer_update to let the user wait.

        Parameters
        ----------
        hidden: bool
            Indicates whether to hide delayed messages.
        """
        base = {"type": 6}
        if hidden:
            base["data"] = {"flags": 64}

        await self.http.post_initial_response(payload=base, data=self.data)
        self.deferred = True
        return

    async def update(
        self,
        content: str | None = MISSING,
        *,
        tts: bool = False,
        embed: discord.Embed = MISSING,
        embeds: list[discord.Embed] = MISSING,
        file: discord.File = MISSING,
        files: list[discord.File] = MISSING,
        hidden: bool = False,
        allowed_mentions: discord.AllowedMentions = None,
        suppress_embeds: bool = False,
        components: list[ActionRow | Button | Selection] = None,
    ):
        """Responds to this interaction by update a message.

        Parameters
        ----------
        content: Optional[str]
            The content of the message to update
        tts: bool
            Indicates if the message should be sent using text-to-speech.
        embed: Optional[discord.Embed]
            The rich embed for the content to update. This cannot be mixed with ``embeds`` parameter.
        embeds: Optional[list[discord.Embed]]
            A list of embeds to update with the content.
            Maximum of 10. This cannot be mixed with the ``embed`` parameter.
        file: Optional[discord.File]
            The file to upload.
        files: Optional[list[discord.File]]
            A list of files to upload. Must be a maximum of 10.
        hidden: bool
            Indicates whether to hide delayed messages.
        allowed_mentions: discord.AllowedMentions
            Controls the mentions being processed in this message.
        suppress_embeds: bool
            Whether to suppress embeds for the message. This updates the message without any embeds if set to True.
        components: list[Components]
            The component to update with the message
        """
        if suppress_embeds or hidden:
            flags = discord.MessageFlags(
                ephemeral=hidden,
                suppress_embeds=(
                    suppress_embeds if suppress_embeds and not self.responded else False
                ),
            )
        else:
            flags = MISSING

        params = handler_message_parameter(
            content=content,
            tts=tts,
            embed=embed,
            embeds=embeds,
            file=file,
            files=files,
            allowed_mentions=allowed_mentions,
            components=components,
            flags=flags,
        )

        if not self.responded:
            if files is not MISSING or file is not MISSING:
                await self.defer_update()

            if self.deferred:
                await self.client.http.edit_message(
                    channel_id=self.channel.id,
                    message_id=self.message.id,
                    params=params,
                )
                self.deferred = False
            else:
                await self.http.post_initial_response(
                    payload={"type": 7, "data": params.payload}, data=self.data
                )
            self.responded = True
        else:
            await self.http.post_followup(
                payload=params.payload,
                form=params.multipart,
                files=params.files,
                data=self.data,
            )


class AutocompleteContext(ApplicationContext):
    """A responded context consisting of an auto complete."""

    def __init__(self, payload: dict, client):
        super().__init__(payload, client)
        self.type = discord.InteractionType.autocomplete

    async def autocomplete(self, choices: list[CommandOptionChoice]):
        """Respond to this interaction by sending an option.

        Parameters
        ----------
        choices: list[CommandOptionChoice]
            choice items that users can choose from in application command option
        """
        self.responded = True
        payload = {
            "type": 8,
            "data": {"choices": [choice.to_dict() for choice in choices]},
        }
        return await self.http.post_initial_response(data=self.data, payload=payload)


class ModalContext(InteractionContext):
    """A responded context consisting of a modal.

    Attributes
    ----------
    custom_id: str
        The ID of modal.
    components: list[TextInput]
        All components that were in the modal
    """

    def __init__(self, payload: dict, client):
        super().__init__(payload, client)
        self.type = discord.InteractionType.modal_submit
        data = payload.get("data", {})
        components = from_payload(data.get("components", []))
        if isinstance(components, list):
            _components = []
            for x in components:
                _components += x.components
        else:
            _components = components
        self.custom_id = data.get("custom_id")
        self.components: list[TextInput] = _components



import re
from setuptools import setup

version = ''
with open('discord/ext/interaction/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('version is not set')


extras_require = {
    'discordpy': ['discord.py'],
    'pycord': ['py-cord'],
    "test": ["pytest", "pytest-cov"],
    "lint": ["pycodestyle", "black"]
}

setup(
    name='Discord-Extension-Interaction',
    version=version,
    packages=['discord.ext.interaction'],
    url='https://github.com/gunyu1019/discord-extension-interaction',
    license='MIT',
    author='gunyu1019',
    extras_require=extras_require,
    author_email='gunyu1019@yhs.kr',
    description='Framework for Application Commands built on discord.py',
    python_requires='>=3.10',
    long_description=open('README.md', encoding='UTF-8').read(),
    long_description_content_type='text/markdown',
    include_package_data=True,
    install_requires=open('requirements.txt', encoding='UTF-8').read(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: Korean',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities'
    ]
)


"""MIT License

Copyright (c) 2021 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import inspect
from typing import Callable, Coroutine, Any

from .commands import (
    ApplicationSubcommand,
    ApplicationSubcommandGroup,
    CommandOption,
    SlashCommand,
    UserCommand,
    ContextMenu,
)
from .utils import async_all


class BaseCore:
    def __init__(self, func: Callable, checks=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.func = func

        if checks is None:
            checks = []
        if hasattr(func, "__commands_checks__"):
            decorator_checks = getattr(func, "__commands_checks__")
            decorator_checks.reverse()
            checks += decorator_checks
        self.checks: list = checks
        self.cog = None

    def __call__(self, *args, **kwargs):
        return self.callback(*args, **kwargs)

    def callback(self, *args, **kwargs) -> Coroutine[Any, Any, Any]:
        if self.cog is None:
            return self.func(*args, **kwargs)
        return self.func(self.cog, *args, **kwargs)

    async def can_run(self, ctx):
        predicates = self.checks
        if len(predicates) == 0:
            # since we have no checks, then we just return True.
            return True

        return await async_all(predicate(ctx) for predicate in predicates)


# Subcommand
class SubCommand(BaseCore, ApplicationSubcommand):
    def __init__(self, func: Callable, parents, checks=None, *args, **kwargs):
        if kwargs.get("name") is None:
            kwargs["name"] = func.__name__
        self.parents: Command | SubCommandGroup = parents
        self.top_parents: Command = kwargs.pop("top_parents", self.parents)
        self.parents.options.append(self)

        options = kwargs.get("options")
        if options is None:
            options = []
        if hasattr(func, "__command_options__"):
            func.__command_options__.reverse()
            options += func.__command_options__
        self.base_options = options

        # kwargs['options'] = get_signature_option(func, options)
        super().__init__(func=func, checks=checks, *args, **kwargs)

    def set_signature_option(self) -> None:
        """Since add_interaction is called and cog information is entered in parents,
        parents value is reflected and skipping_argument value is determined.
        """
        if self.top_parents.cog is not None:
            self.options = get_signature_option(
                self.func, self.base_options, skipping_argument=2
            )
        else:
            self.options = get_signature_option(
                self.func, self.base_options, skipping_argument=1
            )


class SubCommandGroup(BaseCore, ApplicationSubcommandGroup):
    options: list[SubCommand]

    def __init__(self, func: Callable, parents, checks=None, *args, **kwargs):
        if kwargs.get("name") is None:
            kwargs["name"] = func.__name__
        self.parents: Command = parents
        super().__init__(func=func, checks=checks, *args, **kwargs)
        self.parents.options.append(self)

    def subcommand(
        self,
        name: str = None,
        description: str = "No description.",
        cls: classmethod = None,
        checks=None,
        options: list[CommandOption] | None = None,
    ):
        if options is None:
            options = []

        if cls is None:
            cls = SubCommand

        def decorator(func):
            return cls(
                func,
                name=name,
                description=description,
                checks=checks,
                options=options,
                top_parents=self.parents,
                parents=self,
            )

        return decorator

    def set_signature_option(self) -> None:
        """Since add_interaction is called and cog information is entered in parents,
        parents value is reflected and skipping_argument value is determined.
        """
        for opt in self.options:
            opt.set_signature_option()


class BaseCommand(BaseCore):
    def __init__(
        self, func: Callable, checks=None, sync_command: bool = None, *args, **kwargs
    ):
        if kwargs.get("name") is None:
            kwargs["name"] = func.__name__
        super().__init__(func=func, checks=checks, *args, **kwargs)
        self.sync_command: bool = sync_command


class Command(BaseCommand, SlashCommand):
    def __init__(
        self,
        func: Callable,
        checks=None,
        options: list[CommandOption | SubCommand | SubCommandGroup] = None,
        sync_command: bool = None,
        **kwargs
    ):
        if options is None:
            options = []
        if hasattr(func, "__command_options__"):
            func.__command_options__.reverse()
            options += func.__command_options__
        self.base_options = options

        # options = get_signature_option(func, options)
        super().__init__(
            func=func,
            checks=checks,
            sync_command=sync_command,
            options=options,
            **kwargs
        )

    def set_signature_option(self) -> None:
        """Since add_interaction is called and cog information is entered in parents,
        parents value is reflected and skipping_argument value is determined.
        """
        if self.is_subcommand:
            for opt in self.options:
                opt.set_signature_option()
        else:
            if self.cog is not None:
                self.options = get_signature_option(
                    self.func, self.base_options, skipping_argument=2
                )
            else:
                self.options = get_signature_option(
                    self.func, self.base_options, skipping_argument=1
                )
        return

    def subcommand(
        self,
        name: str = None,
        description: str = "No description.",
        cls: classmethod = None,
        checks=None,
        options: list[CommandOption] = None,
    ):
        if options is None:
            options = []

        if cls is None:
            cls = SubCommand

        def decorator(func):
            new_cls = cls(
                func,
                name=name,
                description=description,
                checks=checks,
                options=options,
                top_parents=self,
                parents=self,
            )

            return new_cls

        return decorator

    def subcommand_group(
        self,
        name: str = None,
        description: str = "No description.",
        cls: classmethod = None,
        options: list = None,
    ):
        if options is None:
            options = []

        if cls is None:
            cls = SubCommandGroup

        def decorator(func):
            new_cls = cls(
                func, name=name, description=description, options=options, parents=self
            )
            return new_cls

        return decorator

    @property
    def is_subcommand(self) -> bool:
        for opt in self.options:
            if isinstance(opt, (SubCommand, SubCommandGroup)):
                return True
        else:
            return False


class MemberCommand(BaseCommand, UserCommand):
    pass


class ContextMenuCommand(BaseCommand, ContextMenu):
    pass


decorator_command_types = Command | MemberCommand | ContextMenuCommand


def command(
    name: str = None,
    description: str = "No description.",
    cls: classmethod = None,
    checks=None,
    options: list[CommandOption] = None,
    sync_command: bool = None,
):
    if options is None:
        options = []

    if cls is None:
        cls = Command

    def decorator(func):
        return cls(
            func,
            name=name,
            description=description,
            checks=checks,
            options=options,
            sync_command=sync_command,
        )

    return decorator


def user(
    name: str = None,
    cls: classmethod = None,
    checks=None,
    sync_command: bool = None,
):
    if cls is None:
        cls = MemberCommand

    def decorator(func):
        return cls(func, name=name, checks=checks, sync_command=sync_command)

    return decorator


def context(
    name: str = None,
    cls: classmethod = None,
    checks=None,
    sync_command: bool = None,
):
    if cls is None:
        cls = ContextMenuCommand

    def decorator(func):
        return cls(func, name=name, checks=checks, sync_command=sync_command)

    return decorator


def get_signature_option(func, options, skipping_argument: int = 1):
    signature_arguments = inspect.signature(func).parameters
    arguments = []
    signature_arguments_count = len(signature_arguments) - skipping_argument

    if len(options) == 0 and len(signature_arguments) > skipping_argument - 1:
        for _ in range(signature_arguments_count):
            options.append(CommandOption.empty_option())
    elif signature_arguments_count > len(options):
        for _ in range(signature_arguments_count - len(options)):
            options.append(CommandOption.empty_option())
    elif signature_arguments_count < len(options):
        raise TypeError("number of options and the number of arguments are different.")

    sign_arguments = list(signature_arguments.values())
    for arg in sign_arguments[skipping_argument:]:
        arguments.append(arg)

    for index, opt in enumerate(options):
        options[index].parameter_name = arguments[index].name
        if getattr(opt, "_name") is None:
            setattr(options[index], "_name", arguments[index].name)
        if opt.required or arguments[index].default == arguments[index].empty:
            options[index].required = True
        if getattr(opt, "_type") is None:
            setattr(options[index], "_type", arguments[index].annotation)

        # Check Empty Option
        if arguments[index].annotation is None:
            del options[index]

    return options



"""MIT License

Copyright (c) 2021 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import asyncio
from typing import Any
from collections.abc import Sequence

import discord
from discord.state import ConnectionState
from discord.utils import MISSING

from .components import ActionRow, Button, Selection, from_payload
from .errors import InvalidArgument, AlreadyDeferred
from .http import handler_message_parameter


class Message(discord.Message):
    """Represents a message from Discord for ``discord-extension-interaction``
    This depends on :class:`discord.Message`.

    Attributes
    ----------
    components: list[Component]
        A list of components in the message.
    """

    def __init__(
        self,
        *,
        state: ConnectionState,
        channel: discord.TextChannel | discord.DMChannel | discord.GroupChannel,
        data: dict[str, Any]
    ):
        if "message_reference" in data and "channel_id" not in data.get(
            "message_reference", {}
        ):
            data["message_reference"]["channel_id"] = channel.id
        super().__init__(state=state, channel=channel, data=data)
        self.components = from_payload(data.get("components", []))

    async def send(
        self,
        content: str | None = MISSING,
        *,
        tts: bool = False,
        embed: discord.Embed = MISSING,
        embeds: list[discord.Embed] = MISSING,
        file: discord.File | None = MISSING,
        files: list[discord.File] | None = MISSING,
        allowed_mentions: discord.AllowedMentions = MISSING,
        components: list[ActionRow, Button, Selection] | None = MISSING,
        reference: (
            discord.Message | discord.MessageReference | discord.PartialMessage
        ) = None,
        mention_author: bool = None,
        stickers: list[discord.Sticker, int] | None = MISSING,
        suppress_embeds: bool = False
    ):
        channel = MessageTransferable(state=self._state, channel=self.channel)
        return await channel.send(
            content=content,
            tts=tts,
            embed=embed,
            embeds=embeds,
            file=file,
            files=files,
            allowed_mentions=allowed_mentions,
            components=components,
            reference=reference,
            mention_author=mention_author,
            stickers=stickers,
            suppress_embeds=suppress_embeds,
        )

    async def edit(
        self,
        content: str | None = MISSING,
        *,
        embed: discord.Embed = MISSING,
        embeds: list[discord.Embed] = MISSING,
        attachment: discord.Attachment | discord.File = MISSING,
        attachments: Sequence[discord.Attachment | discord.File] = MISSING,
        allowed_mentions: discord.AllowedMentions = MISSING,
        components: list[ActionRow | Button | Selection] = MISSING,
        stickers: list[discord.Sticker] = MISSING
    ):
        message = MessageEditable(self.channel, self.id)
        return await message.edit(
            content=content,
            embed=embed,
            embeds=embeds,
            attachment=attachment,
            attachments=attachments,
            allowed_mentions=allowed_mentions,
            components=components,
            stickers=stickers,
        )


class MessageCommand(Message):
    """The message command represents the context.

    Notes
    -----
    MessageCommand was officially disabled in v0.1.
    There are cases where this is used via an override.
    However, it is usually not available.
    """

    def __init__(
        self,
        *,
        state: ConnectionState,
        channel: discord.TextChannel | discord.DMChannel | discord.GroupChannel,
        data: dict[str, Any]
    ):
        super().__init__(state=state, channel=channel, data=data)
        options = self.content.split()

        if len(options) >= 1:
            self.name = options[0]
        else:
            self.name = None

        if len(options) >= 2:
            self.options = self.content.split()[1:]
        else:
            self.options = []

        self.deferred = False
        self.deferred_task: asyncio.Task | None = None

    @staticmethod
    def _typing_done_callback(fut: asyncio.Future) -> None:
        # just retrieve any exception and call it a day
        try:
            fut.exception()
        except (asyncio.CancelledError, Exception):
            pass

    async def _do_deferred(self) -> None:
        for count in range(0, 300, 10):
            await super().channel.typing()
            if not self.deferred:
                break
            await asyncio.sleep(10)

    async def defer(self, _: bool = None) -> None:
        if self.deferred:
            raise AlreadyDeferred()
        self.deferred = True
        self.deferred_task = self._state.loop.create_task(self._do_deferred())
        self.deferred_task.add_done_callback(self._typing_done_callback)
        return

    async def send(
        self,
        content=None,
        *,
        tts: bool = False,
        embed: discord.Embed = None,
        embeds: list[discord.Embed] = None,
        file: discord.File = None,
        files: list[discord.File] = None,
        allowed_mentions: discord.AllowedMentions = None,
        components: list[ActionRow | Button | Selection] = None,
        reference: Message | discord.MessageReference | discord.PartialMessage = None,
        mention_author: bool = None,
        stickers: list[discord.Sticker | int] = MISSING,
        suppress_embeds: bool = False
    ) -> Message | None:
        self.deferred = False
        if self.deferred_task is not None:
            if not self.deferred_task.cancelled():
                self.deferred_task.cancel()
        return await super().send(
            content=content,
            tts=tts,
            embed=embed,
            embeds=embeds,
            file=file,
            files=files,
            allowed_mentions=allowed_mentions,
            components=components,
            reference=reference,
            mention_author=mention_author,
            stickers=stickers,
            suppress_embeds=suppress_embeds,
        )


class MessageTransferable:
    """Use :class:`discord.Message` in ``discord-extension-interaction`` to send a message.
    Usage is the same as ``send()`` in :class:`discord.Message`.
    It can use the :class:`Components` in ``discord-extension-interaction``.

    Examples
    ---------
    @client.event
    async def on_message(message):
        state = getattr(client, "_connection")  # For private attributes
        message_send = interaction.MessageSendable(state, message.channel)
        await message_send.send("Hello World")
    """

    def __init__(self, state: ConnectionState, channel):
        self._state = state
        self.channel = channel

    async def send(
        self,
        content: str | None = MISSING,
        *,
        tts: bool = False,
        embed: discord.Embed = MISSING,
        embeds: Sequence[discord.Embed] = MISSING,
        file: discord.File = MISSING,
        files: list[discord.File] = MISSING,
        allowed_mentions: discord.AllowedMentions = MISSING,
        components: list[ActionRow | Button | Selection] = MISSING,
        reference: Message | discord.MessageReference | discord.PartialMessage = None,
        mention_author: bool = None,
        stickers: list[discord.Sticker | int] = MISSING,
        suppress_embeds: bool = False
    ):
        if reference is not None:
            try:
                reference_dict = reference.to_message_reference_dict()
            except AttributeError:
                raise TypeError(
                    "reference parameter must be Message, MessageReference, or PartialMessage"
                ) from None
        else:
            reference_dict = MISSING

        if suppress_embeds:
            flags = discord.MessageFlags(suppress_embeds=suppress_embeds)
        else:
            flags = MISSING

        params = handler_message_parameter(
            content=content,
            tts=tts,
            embed=embed,
            embeds=embeds,
            previous_allowed_mentions=self._state.allowed_mentions,
            file=file,
            files=files,
            allowed_mentions=allowed_mentions,
            flags=flags,
            components=components,
            reference=reference_dict,
            mention_author=mention_author,
            stickers=stickers,
        )

        resp = await self._state.http.send_message(
            params=params, channel_id=self.channel.id
        )
        ret = Message(state=self._state, channel=self.channel, data=resp)
        return ret


class MessageSendable(MessageTransferable):
    """It will be deprecated."""

    pass


class MessageEditable:
    """Use :class:`discord.Message` in ``discord-extension-interaction`` to edit a message.
    Usage is the same as ``edit()`` in :class:`discord.Message`.
    It can use the :class:`Components` in ``discord-extension-interaction``.

    Examples
    ---------
    @client.event
    async def on_message(message):
        state = getattr(client, "_connection")  # For private attributes
        message_send = interaction.MessageEditable(message.channel, message.id)
        await message_send.edit("Hello World")
    """

    def __init__(self, channel, message_id: int):
        self.id = message_id
        self.channel = channel
        self._state = getattr(channel, "_state")

    async def edit(
        self,
        content: str | None = MISSING,
        *,
        embed: discord.Embed = MISSING,
        embeds: list[discord.Embed] = MISSING,
        attachment: discord.Attachment | discord.File = MISSING,
        attachments: Sequence[discord.Attachment | discord.File] = MISSING,
        allowed_mentions: discord.AllowedMentions = MISSING,
        components: list[ActionRow | Button | Selection] = MISSING,
        stickers: list[discord.Sticker] = MISSING
    ):
        if attachment is not MISSING:
            if attachment is not MISSING and attachments is not MISSING:
                raise InvalidArgument()
            attachments = [attachment]
        params = handler_message_parameter(
            content=content,
            embed=embed,
            embeds=embeds,
            previous_allowed_mentions=self._state.allowed_mentions,
            attachments=attachments,
            allowed_mentions=allowed_mentions,
            components=components,
            stickers=stickers,
        )

        await self._state.http.edit_message(
            channel_id=self.channel.id, message_id=self.id, params=params
        )
        return



"""MIT License

Copyright (c) 2021 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import discord
from discord.ext import commands


def load_extension_consider_inheritance(name: str):
    try:
        return getattr(commands, name)
    except (ModuleNotFoundError, AttributeError):
        return getattr(discord, name)


class InvalidArgument(discord.DiscordException):
    """
    Occurs when the Argument value is incorrect.
    Typically, a single item and a variety of items are in the same declaration.
    """

    def __init__(self):
        super(InvalidArgument, self).__init__(
            "Single item and multiple item cannot be used for same function."
        )


class ExtensionFailed(load_extension_consider_inheritance("ExtensionFailed")):
    pass


class NoEntryPointError(load_extension_consider_inheritance("NoEntryPointError")):
    pass


class ExtensionNotFound(load_extension_consider_inheritance("ExtensionNotFound")):
    pass


class ExtensionAlreadyLoaded(
    load_extension_consider_inheritance("ExtensionAlreadyLoaded")
):
    pass


class AlreadyDeferred(Exception):
    """Occurs when a defer is called while a defer is already called."""

    pass


class CheckFailure(commands.CheckFailure):
    """Exception raised when the predicates in :attr:`.Command.checks` have failed.

    This inherits from :exc:`CommandError`
    """

    pass


class CommandRegistrationError(commands.CommandRegistrationError):
    """An exception raised when the command can't be added
    because the name is already taken by a different command.

    This inherits from :exc:`discord.ClientException`
    """

    pass


class CommandNotFound(commands.CommandNotFound):
    """Exception raised when a command is attempted to be invoked
    but no command under that name is found.

    This is not raised for invalid subcommands, rather just the
    initial main command that is attempted to be invoked.

    This inherits from :exc:`CommandError`.
    """

    pass
